# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json
from typing import Any, Optional

from ...domain import actions
from ...query import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
    ConditionType,
    QuerySpecification,
)
from ...sentinels import (
    NotSet,
    is_set,
    value_or_not_set,
)
from ..command import (
    JsonFileOrStrType,
    KeyValuePairsAction,
    RobotoCommand,
    RobotoCommandSet,
)
from ..common_args import (
    ActionTimeoutArg,
    add_action_reference_arg,
    add_compute_requirements_args,
    add_container_parameters_args,
    add_org_arg,
    parse_compute_requirements,
    parse_container_overrides,
)
from ..context import CLIContext

NAME_PARAM_HELP = "The unique name used to reference a trigger."
ACTION_PARAM_HELP = (
    "Partially or fully qualified reference to the action to be triggered."
)


def parse_condition(
    args, context: CLIContext, parser: argparse.ArgumentParser
) -> Optional[ConditionType]:
    if args.condition_json and (args.required_tag or args.required_metadata):
        parser.error(
            "Can only specify '--condition-json' or a combination of '--required-tag' and "
            + "'--required-metadata' expressions, providing both does not work."
        )

    if args.condition_json:
        condition_json: dict[str, Any] = args.condition_json
        if "operator" in condition_json.keys():
            return ConditionGroup.model_validate(condition_json)
        elif "comparator" in condition_json.keys():
            return Condition.model_validate(condition_json)
        else:
            parser.error(
                "Provided '--condition-json' could not be parsed as a Condition or a ConditionGroup."
            )

    if args.required_tag or args.required_metadata:
        conditions = []
        if args.required_tag:
            for tag in args.required_tag:
                conditions.append(
                    Condition(field="tags", comparator=Comparator.Contains, value=tag)
                )

        if args.required_metadata:
            for key, value in args.required_metadata.items():
                conditions.append(
                    Condition(
                        field=f"metadata.{key}",
                        comparator=Comparator.Equals,
                        value=value,
                    )
                )

        return ConditionGroup(operator=ConditionOperator.And, conditions=conditions)

    return None


def parse_overrides(
    args, context: CLIContext, parser: argparse.ArgumentParser
) -> tuple[
    Optional[actions.ComputeRequirements], Optional[actions.ContainerParameters]
]:
    if args.action is None:
        return None, None

    owner_org_id = args.action.owner if args.action.owner else args.org
    action = actions.Action.from_name(
        name=args.action.name,
        digest=args.action.digest,
        owner_org_id=owner_org_id,
        roboto_client=context.roboto_client,
    )

    compute_requirement_overrides = parse_compute_requirements(
        args, action.compute_requirements
    )

    container_overrides = parse_container_overrides(args, action.container_parameters)

    return compute_requirement_overrides, container_overrides


def create(args, context: CLIContext, parser: argparse.ArgumentParser):
    condition = parse_condition(args, context, parser)

    compute_requirement_overrides, container_overrides = parse_overrides(
        args, context, parser
    )

    trigger = actions.Trigger.create(
        name=args.name,
        action_name=args.action.name,
        action_owner_id=args.action.owner,
        action_digest=args.action.digest,
        additional_inputs=args.additional_inputs,
        compute_requirement_overrides=compute_requirement_overrides,
        container_parameter_overrides=container_overrides,
        condition=condition,
        for_each=args.for_each,
        parameter_values=args.parameter_value,
        required_inputs=args.input_data,
        timeout=args.timeout,
        caller_org_id=args.org,
        roboto_client=context.roboto_client,
    )
    print(json.dumps(trigger.to_dict(), indent=2))


def create_setup_parser(parser):
    parser.add_argument("--name", type=str, required=True, help=NAME_PARAM_HELP)
    add_action_reference_arg(
        parser,
        positional=False,
        required=True,
        arg_help=ACTION_PARAM_HELP,
    )
    parser.add_argument(
        "--required-inputs",
        required=True,
        dest="input_data",
        type=str,
        nargs="+",
        action="extend",
        help="""\
        One or many file patterns for data to download from the data source. Examples:
        front camera images, ``--required-inputs '**/cam_front/*.jpg'``;
        front and rear camera images, ``--required-inputs'**/cam_front/*.jpg' --required-inputs '**/cam_rear/*.jpg'``;
        all data, ``--required-inputs '**/*'``.
        """,
    )
    parser.add_argument(
        "--additional-inputs",
        dest="additional_inputs",
        type=str,
        nargs="+",
        action="extend",
        help="""\
        One or many file patterns for data to download from the data source which is NOT considered as part of
        trigger evaluation. Example: front camera images, ``--additional-inputs '**/cam_front/*.jpg'``.""",
    )
    parser.add_argument(
        "--parameter-value",
        required=False,
        metavar="<PARAMETER_NAME>=<PARAMETER_VALUE>",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``<parameter_name>=<parameter_value>`` pairs to pass to the invocation. "
            "``parameter_value`` is parsed as JSON."
        ),
    )
    parser.add_argument(
        "--timeout",
        required=False,
        action=ActionTimeoutArg,
        help="Optional timeout for an action in minutes, defaults to 30 minutes or 12 hours depending on tier.",
    )
    add_org_arg(parser=parser)
    add_compute_requirements_args(parser)
    add_container_parameters_args(parser)
    parser.add_argument(
        "--for-each",
        type=actions.TriggerForEachPrimitive,
        help="Primitive against which this trigger will spawn invocations.",
        default=actions.TriggerForEachPrimitive.Dataset,
        choices=[x.value for x in actions.TriggerForEachPrimitive],
    )
    parser.add_argument(
        "--condition-json",
        type=JsonFileOrStrType,
        help="Inline JSON or a reference to a JSON file "
        + "which expresses a dataset metadata and tag condition which must be met for this trigger "
        + "to run against a given dataset. If provided, 'required-tag' and 'required-metadata' cannot "
        + "be provided.",
    )
    parser.add_argument(
        "--required-tag",
        nargs="*",
        action="extend",
        help="Dataset tags which must be present for "
        + "this trigger to run against a given dataset. If provided, these are used to construct a "
        + "trigger condition, and as such they cannot be used if 'condition-json' is specified.",
    )
    parser.add_argument(
        "--required-metadata",
        nargs="*",
        action=KeyValuePairsAction,
        help="Dataset metadata "
        + "``key=value`` conditions which all must be met for this trigger to run against a given dataset. "
        + "If provided, these are used to construct a trigger condition, and as such they cannot be "
        + "used if ``condition-json`` is specified.",
    )


def update(args, context: CLIContext, parser: argparse.ArgumentParser):
    if args.enabled and args.disabled:
        parser.error("Cannot set both --enabled and --disabled!")

    if (
        args.condition_json or args.required_tag or args.required_metadata
    ) and args.clear_condition:
        parser.error(
            "Cannot specify a condition via '--required-tag', '--required-metadata', or '--condition-json', and "
            + "also clear the condition with '--clear-condition'."
        )

    condition = parse_condition(args, context, parser)

    compute_requirement_overrides, container_overrides = parse_overrides(
        args, context, parser
    )

    trigger = actions.Trigger.from_name(
        name=args.trigger_name,
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    )

    action_name = args.action.name if args.action else None
    action_owner_id = args.action.owner if args.action else None
    action_digest = args.action.digest if args.action else None
    updates: dict[str, Any] = {
        "action_name": value_or_not_set(action_name),
        "action_owner_id": value_or_not_set(action_owner_id),
        "action_digest": value_or_not_set(action_digest),
        "required_inputs": value_or_not_set(args.required_inputs),
        "condition": None if args.clear_condition else value_or_not_set(condition),
        "additional_inputs": value_or_not_set(args.additional_inputs),
        "parameter_values": value_or_not_set(args.parameter_value),
        "compute_requirement_overrides": value_or_not_set(
            compute_requirement_overrides
        ),
        "container_parameter_overrides": value_or_not_set(container_overrides),
        "for_each": value_or_not_set(args.for_each),
        "timeout": args.timeout,
    }

    if args.enabled:
        updates["enabled"] = True
    elif args.disabled:
        updates["enabled"] = False
    else:
        updates["enabled"] = NotSet

    compacted_updates: dict[str, Any] = {k: v for k, v in updates.items() if is_set(v)}
    update_trigger_request = actions.UpdateTriggerRequest.model_validate(
        compacted_updates
    )
    trigger.update(**update_trigger_request.model_dump(exclude_unset=True))
    print(json.dumps(trigger.to_dict(), indent=2))


def update_setup_parser(parser):
    parser.add_argument("trigger_name", type=str, help=NAME_PARAM_HELP)
    add_action_reference_arg(
        parser,
        positional=False,
        required=False,
        arg_help=ACTION_PARAM_HELP,
    )
    parser.add_argument(
        "--required-inputs",
        dest="required_inputs",
        type=str,
        nargs="+",
        action="extend",
        help="""\
        One or many file patterns for data to download from the data source. Examples:
        front camera images, ``--required-inputs '**/cam_front/*.jpg'``;
        front and rear camera images, ``--required-inputs '**/cam_front/*.jpg' --required-inputs '**/cam_rear/*.jpg'``;
        all data, ``--required-inputs '**/*'``.
        """,
    )
    parser.add_argument(
        "--additional-inputs",
        dest="additional_inputs",
        type=str,
        nargs="+",
        action="extend",
        help="""\
        One or many file patterns for data to download from the data source which is NOT considered as part of
        trigger evaluation. Example:
        front camera images, ``--additional-inputs '**/cam_front/*.jpg'``.""",
    )

    parser.add_argument(
        "--timeout",
        required=False,
        action=ActionTimeoutArg,
        type=lambda s: s if s != "null" else None,
        default=NotSet,
        help="Optional timeout for an action in minutes, defaults to 30 minutes or 12 hours depending on tier.",
    )
    add_org_arg(parser=parser)
    add_compute_requirements_args(parser)
    add_container_parameters_args(parser)
    parser.add_argument(
        "--condition-json",
        type=JsonFileOrStrType,
        help="Inline JSON or a reference to a JSON file "
        + "which expresses a dataset metadata and tag condition which must be met for this trigger "
        + "to run against a given dataset. If provided, 'required-tag' and 'required-metadata' cannot "
        + "be provided.",
    )
    parser.add_argument(
        "--required-tag",
        nargs="*",
        action="extend",
        help="Dataset tags which must be present for "
        + "this trigger to run against a given dataset. If provided, these are used to construct a "
        + "trigger condition, and as such they cannot be used if 'condition-json' is specified.",
    )
    parser.add_argument(
        "--required-metadata",
        nargs="*",
        action=KeyValuePairsAction,
        help="Dataset metadata "
        + "key=value conditions which all must be met for this trigger to run against a given dataset. "
        + "If provided, these are used to construct a trigger condition, and as such they cannot be "
        + "used if 'condition-json' is specified.",
    )
    parser.add_argument(
        "--for-each",
        type=actions.TriggerForEachPrimitive,
        help="Primitive against which this trigger will spawn invocations.",
        choices=[x.value for x in actions.TriggerForEachPrimitive],
    )
    parser.add_argument("--enabled", action="store_true", help="Enables this trigger")
    parser.add_argument("--disabled", action="store_true", help="Disables this trigger")
    parser.add_argument(
        "--clear-condition", action="store_true", help="Sets the condition to None"
    )
    parser.add_argument(
        "--parameter-value",
        required=False,
        metavar="<PARAMETER_NAME>=<PARAMETER_VALUE>",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``<parameter_name>=<parameter_value>`` pairs to pass to the invocation. "
            "``parameter_value`` is parsed as JSON. "
        ),
    )


def get(args, context: CLIContext, parser: argparse.ArgumentParser):
    trigger = actions.Trigger.from_name(
        name=args.trigger_name,
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    )
    print(json.dumps(trigger.to_dict(), indent=2))


def get_setup_parser(parser):
    parser.add_argument("trigger_name", type=str, help=NAME_PARAM_HELP)
    add_org_arg(parser=parser)


def search(args, context: CLIContext, parser: argparse.ArgumentParser):
    conditions: list[Condition] = []
    if args.name:
        conditions.append(
            Condition(
                field="name",
                comparator=Comparator.Equals,
                value=args.name,
            )
        )

    if args.action:
        if args.action.name:
            conditions.append(
                Condition(
                    field="action.name",
                    comparator=Comparator.Equals,
                    value=args.action.name,
                )
            )

        if args.action.digest:
            conditions.append(
                Condition(
                    field="action.digest",
                    comparator=Comparator.Equals,
                    value=args.action.digest,
                )
            )

        if args.action.owner:
            conditions.append(
                Condition(
                    field="action.owner",
                    comparator=Comparator.Equals,
                    value=args.action.owner,
                )
            )

    condition = (
        None
        if len(conditions) == 0
        else ConditionGroup(
            conditions=conditions,
            operator=ConditionOperator.And,
        )
    )

    query = QuerySpecification(condition=condition)

    results = actions.Trigger.query(
        query,
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    )
    for trigger in results:
        print(json.dumps(trigger.to_dict(), indent=2))


def search_setup_parser(parser):
    parser.add_argument(
        "--name",
        required=False,
        action="store",
        help="Query by trigger name. Must provide an exact match; patterns are not accepted.",
    )

    add_action_reference_arg(
        parser,
        positional=False,
        required=False,
        arg_help="Query by partially or fully qualified reference to the action to be triggered.",
    )

    add_org_arg(parser=parser)


def delete(args, context: CLIContext, parser: argparse.ArgumentParser):
    trigger = actions.Trigger.from_name(
        name=args.trigger_name,
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    )
    trigger.delete()
    print(f"Successfully deleted trigger '{args.trigger_name}'")


def delete_setup_parser(parser):
    parser.add_argument("trigger_name", type=str, help=NAME_PARAM_HELP)
    add_org_arg(parser=parser)


create_command = RobotoCommand(
    name="create",
    logic=create,
    setup_parser=create_setup_parser,
    command_kwargs={
        "help": "Creates a trigger to automatically invoke an action on datasets when certain criteria are met"
    },
)

get_command = RobotoCommand(
    name="get",
    logic=get,
    setup_parser=get_setup_parser,
    command_kwargs={"help": "Looks up a specific trigger by name"},
)

delete_command = RobotoCommand(
    name="delete",
    logic=delete,
    setup_parser=delete_setup_parser,
    command_kwargs={"help": "Deletes a trigger with a given name"},
)

search_command = RobotoCommand(
    name="search",
    logic=search,
    setup_parser=search_setup_parser,
    command_kwargs={
        "help": "Searches for triggers that match a given condition. Constrained to a single org."
    },
)

update_command = RobotoCommand(
    name="update",
    logic=update,
    setup_parser=update_setup_parser,
    command_kwargs={"help": "Updates one or more fields of an existing trigger."},
)

commands = [create_command, get_command, delete_command, search_command, update_command]

command_set = RobotoCommandSet(
    name="triggers",
    help="Create and edit triggers to automatically invoke actions on datasets.",
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import collections.abc
import json
import os
import pathlib
import typing


def JsonFileOrStrType(arg):
    arg_type = "string"

    if os.path.isfile(arg):
        arg_type = "file"
        with open(arg, "r") as f:
            payload = f.read()
    else:
        payload = arg

    try:
        return json.loads(payload)
    except ValueError:
        raise argparse.ArgumentTypeError(
            "Could not interpret payload {} '{}' as JSON".format(arg_type, arg)
        )


class KeyValuePairsAction(argparse.Action):
    value_dict: dict[str, typing.Any] = {}

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | collections.abc.Sequence[typing.Any] | None,
        option_string: str | None = None,
    ):
        if values is None:
            return

        try:
            for pair in values:
                key, value = pair.split("=")
                if key in self.value_dict:
                    raise parser.error(
                        f"Key '{key}' was defined multiple times for '{self.dest}'"
                    )
                # Attempt to parse the value to better handle numbers, booleans, etc
                parsed_value = value
                try:
                    parsed_value = json.loads(value)
                except json.decoder.JSONDecodeError:
                    pass  # swallow
                self.value_dict[key] = parsed_value

            setattr(namespace, self.dest, self.value_dict)
        except Exception as e:
            raise parser.error(
                f"Failed to parse '{self.dest}' argument '{values}': {e}"
            )


def ExistingPathlibPath(arg):
    path = pathlib.Path(arg)
    if not path.exists():
        raise argparse.ArgumentTypeError(f"Provided path '{arg}' does not exist!")
    return path



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .dataset import Dataset
from .operations import (
    BeginManifestTransactionRequest,
    BeginManifestTransactionResponse,
    BeginSingleFileUploadRequest,
    BeginSingleFileUploadResponse,
    CreateDatasetRequest,
    DeleteDirectoriesRequest,
    QueryDatasetFilesRequest,
    QueryDatasetsRequest,
    RenameDirectoryRequest,
    ReportTransactionProgressRequest,
    UpdateDatasetRequest,
)
from .record import (
    DatasetCredentials,
    DatasetRecord,
)

__all__ = (
    "BeginManifestTransactionRequest",
    "BeginManifestTransactionResponse",
    "BeginSingleFileUploadRequest",
    "BeginSingleFileUploadResponse",
    "CreateDatasetRequest",
    "Dataset",
    "DatasetCredentials",
    "DatasetRecord",
    "QueryDatasetFilesRequest",
    "QueryDatasetsRequest",
    "ReportTransactionProgressRequest",
    "UpdateDatasetRequest",
    "DeleteDirectoriesRequest",
    "RenameDirectoryRequest",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .collection import Collection
from .operations import (
    CreateCollectionRequest,
    UpdateCollectionRequest,
)
from .record import (
    CollectionChangeRecord,
    CollectionChangeSet,
    CollectionContentMode,
    CollectionRecord,
    CollectionResourceRef,
    CollectionResourceType,
)

__all__ = [
    "Collection",
    "CollectionChangeRecord",
    "CollectionChangeSet",
    "CollectionContentMode",
    "CollectionRecord",
    "CollectionResourceRef",
    "CollectionResourceType",
    "CreateCollectionRequest",
    "UpdateCollectionRequest",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import enum
import typing

import pydantic

from .specification import QuerySpecification


class QueryTarget(str, enum.Enum):
    """
    The type of resource a specific query is requesting.
    """

    Collections = "collections"
    Datasets = "datasets"
    Files = "files"
    Topics = "topics"
    TopicMessagePaths = "topic_message_paths"
    Events = "events"


class QueryStatus(enum.Enum):
    """
    The query lifecycle state of a given query.
    """

    Completed = "completed"
    Failed = "failed"
    Scheduled = "scheduled"


class QueryScheme(str, enum.Enum):
    """
    A specific query format/schema which can be used in combination with some context JSON to provide all information
    required to execute a query.
    """

    QuerySpecV1 = "query_spec_v1"
    """
    The initial variant of roboto.query.QuerySpecification which powered search since mid 2023.
    """


class QueryContext(pydantic.BaseModel):
    """
    Context for a query
    """

    query_scheme: QueryScheme
    query: dict[str, typing.Any] = pydantic.Field(default_factory=dict)


class QueryStorageScheme(str, enum.Enum):
    """
    A specific query result storage format/schema which can be used in combination with some context JSON to provide all
    information required to vend query results
    """

    S3ManifestV1 = "s3_manifest_v1"
    """
    Query results are in S3, and a manifest file enumerates the result parts and how to resolve them into rows.
    """


class QueryStorageContext(pydantic.BaseModel):
    """
    Context for query storage
    """

    storage_scheme: QueryStorageScheme
    storage_ctx: dict[str, typing.Any] = pydantic.Field(default_factory=dict)


class QueryRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a query.
    """

    modified: datetime.datetime = pydantic.Field(
        description="The last time the database record for a query was modified."
    )
    org_id: str = pydantic.Field(
        description="The org on behalf of whom a query was run."
    )
    query_id: str = pydantic.Field(
        description="Unique identifier for an individual query."
    )
    query_ctx: QueryContext = pydantic.Field(
        description="The actual query being run and resolved."
    )
    result_count: int = pydantic.Field(
        description="The number of records matched by this query. Defaults to 0 for incomplete queries.",
        default=0,
    )
    status: QueryStatus = pydantic.Field(
        description="The query lifecycle status of this query."
    )
    submitted: datetime.datetime = pydantic.Field(
        description="The time at which this query was first created."
    )
    submitted_by: str = pydantic.Field(description="The user who scheduled this query.")
    target: QueryTarget = pydantic.Field(
        description="The type of data being requested, e.g. 'Datasets' or 'Topics'."
    )


class SubmitStructuredQueryRequest(pydantic.BaseModel):
    """
    Request payload to submit a structured query
    """

    query: QuerySpecification = pydantic.Field(
        description="The conditions, sorting behavior, and limit of this query."
    )
    target: QueryTarget = pydantic.Field(
        description="The type of data being requested, e.g. 'Datasets' or 'Topics'."
    )


class SubmitRoboqlQueryRequest(pydantic.BaseModel):
    """
    Request payload to submit a RoboQL query
    """

    query: typing.Optional[str] = pydantic.Field(
        description="The conditions, sorting behavior, and limit of this query."
    )
    target: QueryTarget = pydantic.Field(
        description="The type of data being requested, e.g. 'Datasets' or 'Topics'."
    )


class SubmitTermQueryRequest(pydantic.BaseModel):
    """
    Request payload to submit a simple term query
    """

    term: str = pydantic.Field(
        default="",
        description="A string search term which this query will attempt to match across any appropriate fields.",
    )
    target: QueryTarget = pydantic.Field(
        description="The type of data being requested, e.g. 'Datasets' or 'Topics'."
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import shlex
import signal
import subprocess
import sys

from ...auth import Permissions
from ...image_registry import ImageRegistry
from ...waiters import TimeoutError, wait_for
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext
from ..terminal import print_error_and_exit


def push(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    inspect_format_string = "{{ .Architecture }}"
    inspect_cmd = (
        f"docker image inspect --format '{inspect_format_string}' {args.local_image}"
    )
    try:
        inspect_completed_process = subprocess.run(
            shlex.split(inspect_cmd),
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError:
        print_error_and_exit(
            f"Could not find locally built image '{args.local_image}'. Is the repository name and tag correct?",
        )

    if inspect_completed_process.stdout.strip() != "amd64":
        print_error_and_exit(
            [
                "Only amd64 images are supported by the Roboto Platform.",
                "You might be seeing this error if running docker on an ARM-based Mac.",
                "Refer to Docker documentation for how to build AMD-compatible images on non-AMD platforms:",
                "https://docs.docker.com/build/building/multi-platform/",
            ]
        )

    image_registry = ImageRegistry(context.roboto_client)
    parts = args.local_image.split(":")
    if len(parts) == 1:
        repo, tag = parts[0], "latest"
    elif len(parts) == 2:
        repo, tag = parts
    else:
        print_error_and_exit("Invalid image format. Expected '<repository>:<tag>'.")
    repository = image_registry.create_repository(repo, org_id=args.org)
    credentials = image_registry.get_temporary_credentials(
        repository["repository_uri"], Permissions.ReadWrite, org_id=args.org
    )
    login_cmd = f"docker login --username {credentials.username} --password-stdin {credentials.registry_url}"
    try:
        subprocess.run(
            shlex.split(login_cmd),
            capture_output=True,
            check=True,
            input=credentials.password,
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        msgs = [
            "Failed to set Docker credentials for Roboto's image registry.",
        ]
        if exc.stdout:
            msgs.append(exc.stdout)
        if exc.stderr:
            msgs.append(exc.stderr)
        print_error_and_exit(msgs)

    image_uri = f"{repository['repository_uri']}:{tag}"
    tag_cmd = f"docker tag {args.local_image} {image_uri}"
    try:
        subprocess.run(
            shlex.split(tag_cmd),
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        msgs = [f"Failed to tag local image '{args.local_image}' as '{image_uri}'."]
        if exc.stdout:
            msgs.append(exc.stdout)
        if exc.stderr:
            msgs.append(exc.stderr)
        print_error_and_exit(msgs)

    push_cmd = f"docker push {image_uri}"
    if args.quiet:
        push_cmd = f"{push_cmd} --quiet"

    with subprocess.Popen(
        shlex.split(push_cmd),
        text=True,
    ) as push_proc:
        try:
            push_proc.wait()
        except KeyboardInterrupt:
            push_proc.kill()
            print("")
            sys.exit(128 + signal.SIGINT.value)

    if not args.quiet:
        print("Waiting for image to be available...")
    try:
        wait_for(
            image_registry.repository_contains_image,
            args=[repository["repository_name"], tag, args.org],
            interval=lambda iteration: min((2**iteration) / 2, 32),
        )
        if not args.quiet:
            print(
                f"Image pushed successfully! You can now use '{image_uri}' in your Roboto Actions."
            )
    except TimeoutError:
        print_error_and_exit(
            "Image could not be confirmed as successfully pushed. Try pushing again in a few minutes."
        )
    except KeyboardInterrupt:
        print("")
        sys.exit(128 + signal.SIGINT.value)


def push_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)

    parser.add_argument(
        "local_image",
        action="store",
        help=(
            "Specify the local image to push, in the format ``<repository>:<tag>``. "
            "If no tag is specified, ``latest`` is assumed. "
            "Image must exist locally (i.e. ``docker images`` must list it)."
        ),
    )

    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="Only output the image URI, terminated by a newline, if successful.",
    )


push_command = RobotoCommand(
    name="push",
    logic=push,
    setup_parser=push_parser,
    command_kwargs={
        "help": (
            "Push a local container image into Roboto's image registry. "
            "Requires Docker CLI."
        )
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import shlex
import subprocess

from ...auth import Permissions
from ...image_registry import ImageRegistry
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def login(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    image_registry = ImageRegistry(context.roboto_client)
    permissions = Permissions(args.permissions)
    credentials = image_registry.get_temporary_credentials(
        args.repository_uri, permissions, org_id=args.org
    )
    cmd = f"docker login --username {credentials.username} --password-stdin {credentials.registry_url}"
    docker_login_completed_process = subprocess.run(
        shlex.split(cmd),
        capture_output=True,
        check=True,
        input=credentials.password,
        text=True,
    )
    print(docker_login_completed_process.stdout)


def login_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)

    parser.add_argument(
        "--repository-uri",
        dest="repository_uri",
        required=True,
        action="store",
        help="Image repository within Roboto's image registry. Login credentials will be scoped to this repository.",
    )

    parser.add_argument(
        "--permissions",
        required=False,
        action="store",
        choices=[p.value for p in Permissions],
        default=Permissions.ReadWrite.value,
        help="Specify the access level for the temporary credentials.",
    )


login_command = RobotoCommand(
    name="login",
    logic=login,
    setup_parser=login_parser,
    command_kwargs={
        "help": (
            "Temporarily login to Roboto's image registry. "
            "Requires Docker CLI. Login is valid for 12 hours and is scoped to a particular repository."
        )
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pydantic


class ContainerImageRepositoryRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a container image repository.
    """

    org_id: str
    repository_name: str
    repository_uri: str
    arn: str


class ContainerImageRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a container image.
    """

    org_id: str
    repository_name: str
    image_tag: str
    image_uri: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import datetime
import typing

from ...http import RobotoClient, StreamedList
from ...query import QuerySpecification
from ...waiters import Interval, wait_for
from .action_record import (
    ComputeRequirements,
    ContainerParameters,
)
from .invocation_operations import (
    SetContainerInfoRequest,
    SetLogsLocationRequest,
    UpdateInvocationStatus,
)
from .invocation_record import (
    ActionProvenance,
    ExecutableProvenance,
    InvocationDataSource,
    InvocationRecord,
    InvocationStatus,
    InvocationStatusRecord,
    LogRecord,
    LogsLocation,
    SourceProvenance,
)


class Invocation:
    """
    An instance of an execution of an action, initiated manually by a user or automatically by a trigger
    """

    __record: InvocationRecord
    __roboto_client: RobotoClient

    @classmethod
    def from_id(
        cls,
        invocation_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Invocation":
        roboto_client = RobotoClient.defaulted(roboto_client)
        response = roboto_client.get(f"v1/actions/invocations/{invocation_id}")
        record = response.to_record(InvocationRecord)
        return cls(record, roboto_client)

    @classmethod
    def query(
        cls,
        spec: typing.Optional[QuerySpecification] = None,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator["Invocation", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        spec = spec or QuerySpecification()

        known = set(InvocationRecord.model_fields.keys())
        actual = set()
        for field in spec.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of Invocation"
                if plural
                else "is not a known attribute of Invocation"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        while True:
            response = roboto_client.post(
                "v1/actions/invocations/query",
                data=spec,
                owner_org_id=owner_org_id,
            )

            paginated_results = response.to_paginated_list(InvocationRecord)
            for record in paginated_results.items:
                yield cls(record, roboto_client)
            if paginated_results.next_token:
                spec.after = paginated_results.next_token
            else:
                break

    def __init__(
        self,
        record: InvocationRecord,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> None:
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def action(self) -> ActionProvenance:
        return self.__record.provenance.action

    @property
    def compute_requirements(self) -> ComputeRequirements:
        return self.__record.compute_requirements

    @property
    def container_parameters(self) -> ContainerParameters:
        return self.__record.container_parameters

    @property
    def created(self) -> datetime.datetime:
        return self.__record.created

    @property
    def current_status(self) -> InvocationStatus:
        sorted_status_records = sorted(
            self.__record.status, key=lambda s: s.status.value
        )
        return sorted_status_records[-1].status

    @property
    def data_source(self) -> InvocationDataSource:
        return self.__record.data_source

    @property
    def executable(self) -> ExecutableProvenance:
        return self.__record.provenance.executable

    @property
    def id(self) -> str:
        return self.__record.invocation_id

    @property
    def input_data(self) -> list[str]:
        return self.__record.input_data

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def parameter_values(self) -> dict[str, typing.Any]:
        return self.__record.parameter_values

    @property
    def reached_terminal_status(self) -> bool:
        return any(
            status_record.status.is_terminal() for status_record in self.__record.status
        )

    @property
    def record(self) -> InvocationRecord:
        return self.__record

    @property
    def source(self) -> SourceProvenance:
        return self.__record.provenance.source

    @property
    def status_log(self) -> list[InvocationStatusRecord]:
        return self.__record.status

    @property
    def timeout(self) -> int:
        return self.__record.timeout

    def cancel(self) -> None:
        if self.current_status.is_terminal():
            return

        self.__roboto_client.post(
            f"v1/actions/invocations/{self.id}/cancel",
            owner_org_id=self.org_id,
        )

    def get_logs(
        self, page_token: typing.Optional[str] = None
    ) -> collections.abc.Generator[LogRecord, None, None]:
        while True:
            response = self.__roboto_client.get(
                f"v1/actions/invocations/{self.id}/logs",
                query={"page_token": page_token} if page_token else None,
                owner_org_id=self.org_id,
            )
            paginated_results = response.to_paginated_list(LogRecord)
            for record in paginated_results.items:
                yield record
            if paginated_results.next_token:
                page_token = paginated_results.next_token
            else:
                break

    def is_queued_for_scheduling(self) -> bool:
        """
        An invocation is queued for scheduling if:
            1. its most recent status is "Queued"
            3. and is not "Deadly"
        """
        if self.current_status != InvocationStatus.Queued:
            return False

        return not any(
            status_record.status == InvocationStatus.Deadly
            for status_record in self.__record.status
        )

    def refresh(self) -> "Invocation":
        self.__record = Invocation.from_id(self.id, self.__roboto_client).record
        return self

    def set_container_image_digest(
        self,
        digest: str,
    ) -> "Invocation":
        """
        This is an admin-only operation to memorialize the digest of the container image
        that was pulled in the course of invoking the action.
        """
        request = SetContainerInfoRequest(image_digest=digest)
        response = self.__roboto_client.put(
            f"v1/actions/invocations/{self.id}/container-info",
            data=request,
            owner_org_id=self.org_id,
        )
        record = response.to_record(InvocationRecord)
        self.__record = record
        return self

    def set_logs_location(self, logs: LogsLocation) -> "Invocation":
        """
        This is an admin-only operation to memorialize the base location where invocation logs are saved.

        Use the "get_logs" or "stream_logs" methods to access invocation logs.
        """
        request = SetLogsLocationRequest(bucket=logs.bucket, prefix=logs.prefix)
        response = self.__roboto_client.put(
            f"v1/actions/invocations/{self.id}/logs",
            data=request,
            owner_org_id=self.org_id,
        )
        record = response.to_record(InvocationRecord)
        self.__record = record
        return self

    def stream_logs(
        self, last_read: typing.Optional[str] = None
    ) -> collections.abc.Generator[LogRecord, None, typing.Optional[str]]:
        while True:
            response = self.__roboto_client.get(
                f"v1/actions/invocations/{self.id}/logs/stream",
                owner_org_id=self.org_id,
                query={"last_read": last_read} if last_read else None,
            )

            response_data = response.to_dict(json_path=["data"])
            streamed_results = StreamedList(
                items=[
                    LogRecord.model_validate(record)
                    for record in response_data["items"]
                ],
                has_next=response_data["has_next"],
                last_read=response_data["last_read"],
            )

            for record in streamed_results.items:
                yield record
            if streamed_results.has_next and streamed_results.last_read:
                last_read = streamed_results.last_read
            else:
                break

        return streamed_results.last_read

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")

    def update_status(
        self, next_status: InvocationStatus, detail: typing.Optional[str] = None
    ) -> "Invocation":
        if next_status == InvocationStatus.Failed:
            # Heuristic: if this is the third time the invocation has failed, it is Deadly
            num_failures = len(
                [
                    status_record
                    for status_record in self.__record.status
                    if status_record.status == InvocationStatus.Failed
                ]
            )
            if num_failures >= 2:
                next_status = InvocationStatus.Deadly

        request = UpdateInvocationStatus(
            status=next_status,
            detail=detail if detail else "None",
        )
        response = self.__roboto_client.post(
            f"v1/actions/invocations/{self.id}/status",
            data=request,
            owner_org_id=self.org_id,
        )
        record = response.to_record(InvocationRecord)
        self.__record = record
        return self

    def wait_for_terminal_status(
        self,
        timeout: float = 60 * 5,
        poll_interval: Interval = 5,
    ) -> None:
        """
        Wait for the invocation to reach a terminal status.

        Throws a :py:exc:`~roboto.waiters.TimeoutError` if the timeout is reached.

        Args:
            timeout: The maximum amount of time, in seconds, to wait for the invocation to reach a terminal status.
            poll_interval: The amount of time, in seconds, to wait between polling iterations.
        """

        def _condition() -> bool:
            self.refresh()
            return self.reached_terminal_status

        return wait_for(
            _condition,
            timeout=timeout,
            interval=poll_interval,
            timeout_msg=f"Timed out waiting for invocation '{self.id}' to reach terminal status",
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import typing
from typing import Any, Optional

import pydantic

from ...time import utcnow
from ..files import S3Credentials


class DatasetCredentials(pydantic.BaseModel):
    """
    Handle credentials for dataset file access
    """

    access_key_id: str
    bucket: str
    expiration: datetime.datetime
    secret_access_key: str
    session_token: str
    region: str
    required_prefix: str

    def is_expired(self) -> bool:
        return utcnow() >= self.expiration

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(mode="json")

    def to_s3_credentials(self) -> S3Credentials:
        return {
            "access_key": self.access_key_id,
            "secret_key": self.secret_access_key,
            "token": self.session_token,
            "expiry_time": self.expiration.isoformat(),
            "region": self.region,
        }


def make_backwards_compatible_placeholder_storage_ctx() -> dict[str, typing.Any]:
    """
    Because of some overly aggressive pydantic model validation, we need to return a storage_context with our
    original S3 description in order to stop SDK clients prior to 0.10.0 from throwing errors.
    """
    return {
        "bucket_name": "NOT_SET",
        "iam_role_arn": "NOT_SET",
        "key_prefix": "NOT_SET",
    }


class DatasetRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a dataset.
    """

    created: datetime.datetime
    created_by: str
    dataset_id: str  # sort key
    description: Optional[str] = None
    device_id: Optional[str] = None
    metadata: dict[str, Any] = pydantic.Field(default_factory=dict)
    modified: datetime.datetime
    modified_by: str
    org_id: str  # partition key
    roboto_record_version: int = 0  # A protected field, incremented on every update
    tags: list[str] = pydantic.Field(default_factory=list)

    # Because datasets may have files in many buckets, both customer provided and Roboto managed, having a single
    # storage location or a single administrator no longer makes sense.
    #
    # These fields are deprecated, and have defaulting strategies to maintain backwards compatibility with old
    # versions of the Roboto SDK, since this record defines the shape of the service's return payload.
    administrator: str = "Roboto"
    storage_ctx: dict[str, Any] = pydantic.Field(
        default_factory=make_backwards_compatible_placeholder_storage_ctx
    )
    storage_location: str = "S3"



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import copy
import typing

import pydantic

StrSequence = typing.Union[list[str], tuple[str, ...], set[str]]


# TODO: consider retiring this type in favor of MetadataChangeset.
class TaglessMetadataChangeset(pydantic.BaseModel):
    """Changeset for tagless metadata"""

    # Add each field in this dict if it doesn't exist, else overwrite the existing value
    # Expands dot notation to nested objects
    put_fields: typing.Optional[dict[str, typing.Any]] = None

    # Remove each field in this sequence if it exists
    # Expands dot notation to nested objects
    remove_fields: typing.Optional[StrSequence] = None

    # Replace all existing metadata with the fields provided in 'put_fields'.
    # Set by appropriate class methods.
    replace_all: bool = False

    class Builder:
        __put_fields: dict[str, typing.Any]
        __remove_fields: list[str]

        def __init__(self) -> None:
            self.__put_fields = dict()
            self.__remove_fields = []

        def put_field(
            self, key: str, value: typing.Any
        ) -> "TaglessMetadataChangeset.Builder":
            self.__put_fields[key] = value
            return self

        def remove_field(self, key: str) -> "TaglessMetadataChangeset.Builder":
            self.__remove_fields.append(key)
            return self

        def build(self) -> "TaglessMetadataChangeset":
            changeset: collections.abc.Mapping = {
                "put_fields": self.__put_fields,
                "remove_fields": self.__remove_fields,
            }
            return TaglessMetadataChangeset(**{k: v for k, v in changeset.items() if v})

    @classmethod
    def from_replacement_metadata(
        cls, metadata: dict[str, typing.Any]
    ) -> "TaglessMetadataChangeset":
        """Creates a changeset to replace any existing metadata with the metadata provided."""

        return cls(put_fields=metadata, replace_all=True)

    def apply_field_updates(
        self, existing_metadata: dict[str, typing.Any]
    ) -> dict[str, typing.Any]:
        if self.replace_all:
            return self.put_fields or {}

        updated_metadata = copy.deepcopy(existing_metadata)
        if self.put_fields:
            for key, value in self.put_fields.items():
                self.__set_nested(updated_metadata, key, value)
        if self.remove_fields:
            for key in self.remove_fields:
                self.__del_nested(updated_metadata, key)
        return updated_metadata

    def has_changes(self) -> bool:
        """Checks whether applying this changeset would result in any metadata changes."""

        return self.replace_all or any(
            [len(self.put_fields or {}), len(self.remove_fields or [])]
        )

    def __set_nested(
        self, obj: dict[str, typing.Any], key: str, value: typing.Any
    ) -> None:
        """
        Set nested path to value using dot notation
        """
        keys = key.split(".")
        subobj = obj
        for key in keys[:-1]:
            if isinstance(subobj, collections.abc.MutableSequence):
                key = int(key)
                if key >= len(subobj):
                    raise IndexError(f"Index {key} out of range for {subobj}")
                subobj = subobj[key]
            else:
                subobj = subobj.setdefault(key, {})

        if isinstance(subobj, collections.abc.MutableMapping):
            subobj[keys[-1]] = value
        elif isinstance(subobj, collections.abc.MutableSequence):
            subobj.insert(int(keys[-1]), value)

    def __del_nested(self, obj: dict[str, typing.Any], key: str) -> None:
        """
        Delete a value from nested path using dot notation
        """

        def __key_in_collection(
            key: typing.Union[str, int], collection: collections.abc.Collection
        ) -> bool:
            if isinstance(collection, collections.abc.MutableSequence):
                key = int(key)
                return key < len(collection)
            else:
                return key in collection

        def __del_from_collection(
            key: typing.Union[str, int], collection: collections.abc.Collection
        ) -> None:
            if not __key_in_collection(key, collection):
                return

            if isinstance(collection, collections.abc.MutableMapping):
                del collection[key]
            elif isinstance(collection, collections.abc.MutableSequence):
                collection.pop(int(key))

        keys = key.split(".")
        path: list[
            tuple[
                # subobj
                typing.Union[
                    collections.abc.MutableMapping, collections.abc.MutableMapping
                ],
                # key
                typing.Optional[typing.Union[str, int]],
                # parent_obj
                typing.Optional[
                    typing.Union[
                        collections.abc.MutableMapping, collections.abc.MutableMapping
                    ]
                ],
            ]
        ] = [(obj, None, None)]
        for key in keys[:-1]:
            sub_obj, _, _ = path[-1]
            if isinstance(sub_obj, collections.abc.MutableSequence):
                key = int(key)

            if not __key_in_collection(key, sub_obj):
                return

            path.append((sub_obj[key], key, sub_obj))

        __del_from_collection(keys[-1], path[-1][0])

        # remove any now empty collections, working backwards
        for _, key_path, parent_obj in reversed(path):
            if key_path is None or parent_obj is None:
                # root object
                return

            if isinstance(parent_obj, collections.abc.MutableSequence):
                key_path = int(key_path)

            if not __key_in_collection(key_path, parent_obj):
                return

            if not len(parent_obj[key_path]):
                __del_from_collection(key_path, parent_obj)


# TODO https://roboto.atlassian.net/browse/ROBO-903, change this to a mixin model to handle tags and/or metadata
class MetadataChangeset(pydantic.BaseModel):
    """Changeset for metadata"""

    # Add each tag in this sequence if it doesn't exist
    put_tags: typing.Optional[StrSequence] = None
    # Remove each tag in this sequence if it exists
    remove_tags: typing.Optional[StrSequence] = None
    # Add each field in this dict if it doesn't exist, else overwrite the existing value
    # Expands dot notation to nested objects
    put_fields: typing.Optional[dict[str, typing.Any]] = None
    # Remove each field in this sequence if it exists
    # Expands dot notation to nested objects
    remove_fields: typing.Optional[StrSequence] = None

    # Controls whether the changeset will act as a wholesale replacement of existing metadata or tags.
    # Intended for use with dedicated class factories only.
    replace_all: bool = False

    class Builder:
        __put_tags: list[str]
        __remove_tags: list[str]
        __put_fields: dict[str, typing.Any]
        __remove_fields: list[str]

        def __init__(self) -> None:
            self.__put_tags = []
            self.__remove_tags = []
            self.__put_fields = dict()
            self.__remove_fields = []

        def put_tag(self, tag: str) -> "MetadataChangeset.Builder":
            self.__put_tags.append(tag)
            return self

        def remove_tag(self, tag: str) -> "MetadataChangeset.Builder":
            self.__remove_tags.append(tag)
            return self

        def put_field(self, key: str, value: typing.Any) -> "MetadataChangeset.Builder":
            self.__put_fields[key] = value
            return self

        def remove_field(self, key: str) -> "MetadataChangeset.Builder":
            self.__remove_fields.append(key)
            return self

        def build(self) -> "MetadataChangeset":
            changeset: collections.abc.Mapping = {
                "put_tags": self.__put_tags,
                "remove_tags": self.__remove_tags,
                "put_fields": self.__put_fields,
                "remove_fields": self.__remove_fields,
            }
            return MetadataChangeset(**{k: v for k, v in changeset.items() if v})

    @classmethod
    def from_metadata(
        cls, metadata: typing.Mapping[str, typing.Any]
    ) -> "MetadataChangeset":
        """Creates a changeset that will replace any existing metadata with what's provided."""
        return cls(put_fields=dict(metadata), replace_all=True)

    def apply_field_updates(
        self, existing_metadata: dict[str, typing.Any]
    ) -> dict[str, typing.Any]:
        if self.replace_all:
            return self.put_fields or {}  # should we deep-copy?

        updated_metadata = copy.deepcopy(existing_metadata)
        if self.put_fields:
            for key, value in self.put_fields.items():
                self.__set_nested(updated_metadata, key, value)
        if self.remove_fields:
            for key in self.remove_fields:
                self.__del_nested(updated_metadata, key)
        return updated_metadata

    def apply_tag_updates(self, existing_tags: list[str]) -> StrSequence:
        if self.replace_all:
            return self.put_tags or []  # should we copy?

        updated_tags = existing_tags.copy()
        if self.put_tags:
            for tag in self.put_tags:
                if tag not in updated_tags:
                    updated_tags.append(tag)
        if self.remove_tags:
            updated_tags = [tag for tag in updated_tags if tag not in self.remove_tags]

        return updated_tags

    @staticmethod
    def __combine_strseq_field(
        first: typing.Optional[StrSequence], second: typing.Optional[StrSequence]
    ) -> typing.Optional[StrSequence]:
        combined = list(set(first or []) | set(second or []))
        return None if len(combined) == 0 else combined

    @staticmethod
    def __combine_dict_field(
        first: typing.Optional[dict[str, typing.Any]],
        second: typing.Optional[dict[str, typing.Any]],
    ) -> typing.Optional[dict[str, typing.Any]]:
        combined: dict[str, typing.Any] = dict()
        combined.update(first or {})
        combined.update(second or {})
        return None if len(combined) == 0 else combined

    def combine(self, other: "MetadataChangeset") -> "MetadataChangeset":
        return MetadataChangeset(
            put_tags=self.__combine_strseq_field(self.put_tags, other.put_tags),
            remove_tags=self.__combine_strseq_field(
                self.remove_tags, other.remove_tags
            ),
            remove_fields=self.__combine_strseq_field(
                self.remove_fields, other.remove_fields
            ),
            put_fields=self.__combine_dict_field(self.put_fields, other.put_fields),
        )

    def is_empty(self) -> bool:
        return not self.replace_all and not any(
            [
                self.put_tags,
                self.remove_tags,
                self.put_fields,
                self.remove_fields,
            ]
        )

    def __set_nested(
        self, obj: dict[str, typing.Any], key: str, value: typing.Any
    ) -> None:
        """
        Set nested path to value using dot notation
        """
        keys = key.split(".")
        subobj = obj
        for key in keys[:-1]:
            if isinstance(subobj, collections.abc.MutableSequence):
                key = int(key)
                if key >= len(subobj):
                    raise IndexError(f"Index {key} out of range for {subobj}")
                subobj = subobj[key]
            else:
                subobj = subobj.setdefault(key, {})

        if isinstance(subobj, collections.abc.MutableMapping):
            subobj[keys[-1]] = value
        elif isinstance(subobj, collections.abc.MutableSequence):
            subobj.insert(int(keys[-1]), value)

    def __del_nested(self, obj: dict[str, typing.Any], key: str) -> None:
        """
        Delete a value from nested path using dot notation
        """

        def __key_in_collection(
            key: typing.Union[str, int], collection: collections.abc.Collection
        ) -> bool:
            if isinstance(collection, collections.abc.MutableSequence):
                key = int(key)
                return key < len(collection)
            else:
                return key in collection

        def __del_from_collection(
            key: typing.Union[str, int], collection: collections.abc.Collection
        ) -> None:
            if not __key_in_collection(key, collection):
                return

            if isinstance(collection, collections.abc.MutableMapping):
                del collection[key]
            elif isinstance(collection, collections.abc.MutableSequence):
                collection.pop(int(key))

        keys = key.split(".")
        path: list[
            tuple[
                # subobj
                typing.Union[
                    collections.abc.MutableMapping, collections.abc.MutableMapping
                ],
                # key
                typing.Optional[typing.Union[str, int]],
                # parent_obj
                typing.Optional[
                    typing.Union[
                        collections.abc.MutableMapping, collections.abc.MutableMapping
                    ]
                ],
            ]
        ] = [(obj, None, None)]
        for key in keys[:-1]:
            sub_obj, _, _ = path[-1]
            if isinstance(sub_obj, collections.abc.MutableSequence):
                key = int(key)

            if not __key_in_collection(key, sub_obj):
                return

            path.append((sub_obj[key], key, sub_obj))

        __del_from_collection(keys[-1], path[-1][0])

        # remove any now empty collections, working backwards
        for _, key_path, parent_obj in reversed(path):
            if key_path is None or parent_obj is None:
                # root object
                return

            if isinstance(parent_obj, collections.abc.MutableSequence):
                key_path = int(key_path)

            if not __key_in_collection(key_path, parent_obj):
                return

            if not len(parent_obj[key_path]):
                __del_from_collection(key_path, parent_obj)


class UpdateCondition(pydantic.BaseModel):
    """
    A condition to be applied to an update operation, succeeding only if the condition evaluates to True at update-time.

    `value` is compared to the resource's current value of `key` using `comparator`.

    This is a severely constrainted subset of the conditions supported by DynamoDB. See:
    https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.OperatorsAndFunctions.html
    """

    key: str
    value: typing.Any = None
    # Comparators are tied to convenience methods exposed on boto3.dynamodb.conditions.Attr. See:
    # https://github.com/boto/boto3/blob/5ad1a624111ed25efc81f425113fa51150516bb4/boto3/dynamodb/conditions.py#L246
    comparator: typing.Literal["eq", "ne"]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import json
import typing

import boto3
import botocore.exceptions
import pydantic

from ...exceptions import (
    RobotoInvalidRequestException,
)
from ...http import RobotoClient
from ...logging import default_logger

logger = default_logger()


class RegisterS3IntegrationRequest(pydantic.BaseModel):
    """
    Request payload to integrate an S3 bucket
    """

    account_id: str
    aws_region: str
    bucket_name: str
    org_id: str
    transfer_accelerated: bool = False
    readonly: bool = False


class RegisterS3IntegrationResponse(pydantic.BaseModel):
    """
    Response payload to integrate an S3 bucket
    """

    iam_role_name: str
    iam_role_policy: dict[str, typing.Any]
    iam_role_trust_relationship: dict[str, typing.Any]
    s3_bucket_cors_policy: list[dict[str, typing.Any]]


class S3IntegrationService:
    __roboto_client: RobotoClient
    __s3_client: typing.Any  # boto3.client("s3")
    __sts_client: typing.Any  # boto3.client("sts")
    __iam_client: typing.Any  # boto3.client("iam")

    def __init__(
        self,
        roboto_client: RobotoClient,
        sts_client: typing.Optional[typing.Any] = None,
        s3_client: typing.Optional[typing.Any] = None,
        iam_client: typing.Optional[typing.Any] = None,
    ):
        self.__s3_client = s3_client or boto3.client("s3")
        self.__sts_client = sts_client or boto3.client("sts")
        self.__iam_client = iam_client or boto3.client("iam")
        self.__roboto_client = roboto_client

    def register_bucket(
        self,
        org_id: str,
        account_id: str,
        bucket_name: str,
        transfer_accelerated: bool = False,
        readonly: bool = False,
    ):
        logger.info("Checking that you have access to the specified account and bucket")

        try:
            available_account_id = self.__sts_client.get_caller_identity().get(
                "Account"
            )
        except botocore.exceptions.ClientError:
            raise RobotoInvalidRequestException("Couldn't get caller ID from AWS")

        if available_account_id != account_id:
            raise RobotoInvalidRequestException(
                f"Account ID mismatch: you're trying to add a bucket to account {account_id}, "
                + f"but your current credentials are for {available_account_id}"
            )

        buckets = self.__s3_client.list_buckets().get("Buckets", [])
        bucket_names = {bucket["Name"] for bucket in buckets}

        if bucket_name not in bucket_names:
            raise ValueError(f"Account {account_id} does not own bucket {bucket_name}")

        aws_region = self.__s3_client.get_bucket_location(Bucket=bucket_name)[
            "LocationConstraint"
        ]

        response = self.__roboto_client.post(
            "v1/integrations/storage/s3/register",
            data=RegisterS3IntegrationRequest(
                account_id=account_id,
                bucket_name=bucket_name,
                org_id=org_id,
                aws_region=aws_region,
                transfer_accelerated=transfer_accelerated,
                readonly=readonly,
            ),
        ).to_record(RegisterS3IntegrationResponse)

        self.__iam_client.create_role(
            RoleName=response.iam_role_name,
            AssumeRolePolicyDocument=json.dumps(response.iam_role_trust_relationship),
            Description=f"Cross-account role which grants Roboto read/write access to the bucket '{bucket_name}'.",
        )

        self.__iam_client.put_role_policy(
            RoleName=response.iam_role_name,
            # Policy name limit is 128 characters, we'll cap at a couple characters short of that.
            PolicyName=f"RobotoS3Access-{bucket_name[:110]}",
            PolicyDocument=json.dumps(response.iam_role_policy),
        )

        self.__s3_client.put_bucket_cors(
            Bucket=bucket_name,
            CORSConfiguration={"CORSRules": response.s3_bucket_cors_policy},
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...image_registry import ImageRegistry
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def delete_image(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    image_registry = ImageRegistry(context.roboto_client)
    image_registry.delete_image(args.remote_image, org_id=args.org)


def delete_image_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)

    parser.add_argument(
        "remote_image",
        action="store",
        help="Specify the remote image to delete, in the format ``<repository>:<tag>``.",
    )


delete_image_command = RobotoCommand(
    name="delete-image",
    logic=delete_image,
    setup_parser=delete_image_parser,
    command_kwargs={
        "help": (
            "Delete a container image hosted in Roboto's image registry. "
            "Requires Docker CLI."
        )
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import functools
import importlib.metadata
import math
import os
import pathlib
import threading
import typing

import pathspec

from ...association import Association
from ...auth import Permissions
from ...env import RobotoEnv
from ...exceptions import RobotoInternalException
from ...http import PaginatedList, RobotoClient
from ...logging import default_logger
from ...paths import excludespec_from_patterns
from ...query import QuerySpecification
from ...updates import (
    MetadataChangeset,
    StrSequence,
    UpdateCondition,
)
from ..files import (
    DirectoryRecord,
    File,
    FileRecord,
)
from ..files.file_service import FileService
from ..files.progress import (
    TqdmProgressMonitorFactory,
)
from ..topics import Topic
from .operations import (
    BeginManifestTransactionRequest,
    BeginManifestTransactionResponse,
    CreateDatasetRequest,
    QueryDatasetFilesRequest,
    RenameDirectoryRequest,
    ReportTransactionProgressRequest,
    UpdateDatasetRequest,
)
from .record import (
    DatasetCredentials,
    DatasetRecord,
)

logger = default_logger()


MAX_FILES_PER_MANIFEST = 500


class Dataset:
    """
    Contains files in a directory structure.

    Typically, a dataset stores the files from a single robot activity, such as a drone flight
    or an autonomous vehicle mission, however, it is versatile enough to be a
    general-purpose assembly of files.

    """

    UPLOAD_REPORTING_BATCH_COUNT: typing.ClassVar[int] = 10
    """
    Number of batches to break a large upload into for the purpose of reporting progress.
    """
    UPLOAD_REPORTING_MIN_BATCH_SIZE: typing.ClassVar[int] = 10
    """
    Minimum number of files that must be uploaded before reporting progress.
    """

    __roboto_client: RobotoClient
    __record: DatasetRecord
    __transaction_completed_unreported_items: dict[str, set[str]]
    __transaction_completed_mutex: threading.Lock

    @classmethod
    def create(
        cls,
        description: typing.Optional[str] = None,
        metadata: typing.Optional[dict[str, typing.Any]] = None,
        tags: typing.Optional[list[str]] = None,
        caller_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Dataset":
        roboto_client = RobotoClient.defaulted(roboto_client)
        request = CreateDatasetRequest(
            description=description, metadata=metadata or {}, tags=tags or []
        )
        record = roboto_client.post(
            "v1/datasets", data=request, caller_org_id=caller_org_id
        ).to_record(DatasetRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def from_id(
        cls, dataset_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> "Dataset":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(f"v1/datasets/{dataset_id}").to_record(DatasetRecord)
        return cls(record, roboto_client)

    @classmethod
    def query(
        cls,
        spec: typing.Optional[QuerySpecification] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
        owner_org_id: typing.Optional[str] = None,
    ) -> collections.abc.Generator["Dataset", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        spec = spec if spec is not None else QuerySpecification()

        known = set(DatasetRecord.model_fields.keys())
        actual = set()
        for field in spec.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of Dataset"
                if plural
                else "is not a known attribute of Dataset"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        while True:
            paginated_results = roboto_client.post(
                "v1/datasets/query",
                data=spec,
                owner_org_id=owner_org_id,
                idempotent=True,
            ).to_paginated_list(DatasetRecord)

            for record in paginated_results.items:
                yield cls(record, roboto_client)

            if paginated_results.next_token:
                spec.after = paginated_results.next_token
            else:
                break

    def __init__(
        self, record: DatasetRecord, roboto_client: typing.Optional[RobotoClient] = None
    ) -> None:
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__record = record
        self.__transaction_completed_unreported_items = {}
        self.__transaction_completed_mutex = threading.Lock()

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def dataset_id(self) -> str:
        return self.__record.dataset_id

    @property
    def metadata(self) -> dict[str, typing.Any]:
        return self.__record.metadata.copy()

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def record(self) -> DatasetRecord:
        return self.__record

    @property
    def tags(self) -> list[str]:
        return self.__record.tags.copy()

    def delete(self) -> None:
        self.__roboto_client.delete(f"v1/datasets/{self.dataset_id}")

    def delete_files(
        self,
        include_patterns: typing.Optional[list[str]] = None,
        exclude_patterns: typing.Optional[list[str]] = None,
    ) -> None:
        """
        Delete files associated with this dataset.

        `include_patterns` and `exclude_patterns` are lists of gitignore-style patterns.
        See https://git-scm.com/docs/gitignore.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.delete_files(
            ...    include_patterns=["**/*.png"],
            ...    exclude_patterns=["**/back_camera/**"]
            ... )
        """
        for file in self.list_files(include_patterns, exclude_patterns):
            file.delete()

    def download_files(
        self,
        out_path: pathlib.Path,
        include_patterns: typing.Optional[list[str]] = None,
        exclude_patterns: typing.Optional[list[str]] = None,
    ) -> None:
        """
        Download files associated with this dataset to the given directory.
        If `out_path` does not exist, it will be created.

        `include_patterns` and `exclude_patterns` are lists of gitignore-style patterns.
        See https://git-scm.com/docs/gitignore.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.download_files(
            ...     pathlib.Path("/tmp/tmp.nV1gdW5WHV"),
            ...     include_patterns=["**/*.g4"],
            ...     exclude_patterns=["**/test/**"]
            ... )
        """
        if not out_path.is_dir():
            out_path.mkdir(parents=True)

        def _file_to_download_tuple(f: File) -> tuple[FileRecord, pathlib.Path]:
            return f.record, out_path / f.relative_path

        all_files = list(self.list_files(include_patterns, exclude_patterns))

        files_by_bucket = collections.defaultdict(list)
        for file in all_files:
            files_by_bucket[file.record.bucket].append(file)

        for bucket_name, bucket_files in files_by_bucket.items():

            def _file_generator():
                for x in map(
                    _file_to_download_tuple,
                    bucket_files,
                ):
                    yield x

            self.__file_service().download_files(
                file_generator=_file_generator(),
                credential_provider=self.__download_creds_provider(bucket_name),
                progress_monitor_factory=TqdmProgressMonitorFactory(
                    concurrency=1,
                    ctx={
                        "base_path": self.dataset_id,
                        "total_file_count": len(all_files),
                    },
                ),
                max_concurrency=8,
            )

    def get_file_by_path(
        self,
        relative_path: typing.Union[str, pathlib.Path],
        version_id: typing.Optional[int] = None,
    ) -> File:
        """
        Get a File object for the given relative path.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> file = dataset.get_file_by_path("foo/bar.txt")
            >>> print(file.file_id)
            file-abc123
        """
        return File.from_path_and_dataset_id(
            file_path=relative_path,
            dataset_id=self.dataset_id,
            version_id=version_id,
            roboto_client=self.__roboto_client,
        )

    def get_topics(self) -> collections.abc.Generator[Topic, None, None]:
        """
        List all topics associated with files in this dataset. If multiple files have topics with the same name (i.e.
        if a dataset has chunked files with the same schema), they'll be returned as separate topic objects.
        """
        return Topic.get_by_dataset(self.dataset_id, self.__roboto_client)

    def get_topics_by_file(
        self, relative_path: typing.Union[str, pathlib.Path]
    ) -> collections.abc.Generator[Topic, None, None]:
        file = self.get_file_by_path(relative_path)
        return file.get_topics()

    def list_files(
        self,
        include_patterns: typing.Optional[list[str]] = None,
        exclude_patterns: typing.Optional[list[str]] = None,
    ) -> collections.abc.Generator[File, None, None]:
        """
        List files associated with this dataset.

        `include_patterns` and `exclude_patterns` are lists of gitignore-style patterns.
        See https://git-scm.com/docs/gitignore.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> for file in dataset.list_files(
            ...     include_patterns=["**/*.g4"],
            ...     exclude_patterns=["**/test/**"]
            ... ):
            ...     print(file.relative_path)
        """

        page_token: typing.Optional[str] = None
        while True:
            paginated_results = self.__list_files_page(
                page_token=page_token,
                include_patterns=include_patterns,
                exclude_patterns=exclude_patterns,
            )
            for record in paginated_results.items:
                yield File(record, self.__roboto_client)
            if paginated_results.next_token:
                page_token = paginated_results.next_token
            else:
                break

    def put_metadata(
        self,
        metadata: dict[str, typing.Any],
    ) -> None:
        """
        Set each `key`: `value` in this dict as dataset metadata if it doesn't exist, else overwrite the existing value.
        Keys must be strings. Dot notation is supported for nested keys.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.put_metadata({
            ...     "foo": "bar",
            ...     "baz.qux": 101,
            ... })

        """
        self.update(metadata_changeset=MetadataChangeset(put_fields=metadata))

    def put_tags(
        self,
        tags: StrSequence,
    ) -> None:
        """Add each tag in this sequence if it doesn't exist"""
        self.update(
            metadata_changeset=MetadataChangeset(put_tags=tags),
        )

    def refresh(self) -> "Dataset":
        """Refresh the underlying dataset record with the latest server state."""
        self.__record = self.__roboto_client.get(
            f"v1/datasets/{self.dataset_id}"
        ).to_record(DatasetRecord)
        return self

    def remove_metadata(
        self,
        metadata: StrSequence,
    ) -> None:
        """
        Remove each key in this sequence from dataset metadata if it exists.
        Keys must be strings. Dot notation is supported for nested keys.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.remove_metadata(["foo", "baz.qux"])
        """
        self.update(metadata_changeset=MetadataChangeset(remove_fields=metadata))

    def remove_tags(
        self,
        tags: StrSequence,
    ) -> None:
        """Remove each tag in this sequence if it exists"""
        self.update(metadata_changeset=MetadataChangeset(remove_tags=tags))

    def to_association(self) -> Association:
        return Association.dataset(self.dataset_id)

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")

    def update(
        self,
        conditions: typing.Optional[list[UpdateCondition]] = None,
        description: typing.Optional[str] = None,
        metadata_changeset: typing.Optional[MetadataChangeset] = None,
    ) -> "Dataset":
        request = UpdateDatasetRequest(
            conditions=conditions,
            description=description,
            metadata_changeset=metadata_changeset,
        )

        self.__record = self.__roboto_client.put(
            f"/v1/datasets/{self.dataset_id}",
            data=request,
        ).to_record(DatasetRecord)
        return self

    def rename_directory(self, old_path: str, new_path: str) -> DirectoryRecord:
        response = self.__roboto_client.put(
            f"v1/datasets/{self.dataset_id}/directory/rename",
            data=RenameDirectoryRequest(
                old_path=old_path,
                new_path=new_path,
            ),
        )

        return response.to_record(DirectoryRecord)

    def upload_directory(
        self,
        directory_path: pathlib.Path,
        include_patterns: typing.Optional[list[str]] = None,
        exclude_patterns: typing.Optional[list[str]] = None,
        delete_after_upload: bool = False,
        max_batch_size: int = MAX_FILES_PER_MANIFEST,
    ) -> None:
        """
        Uploads all files and directories recursively from the specified directory path. You can use
        `include_patterns` and `exclude_patterns` to control what files and directories are uploaded, and can
        use `delete_after_upload` to clean up your local filesystem after the uploads succeed.

        Example:
            >>> from roboto import Dataset
            >>> dataset = Dataset(...)
            >>> dataset.upload_directory(
            ...     pathlib.Path("/path/to/directory"),
            ...     exclude_patterns=[
            ...         "__pycache__/",
            ...         "*.pyc",
            ...         "node_modules/",
            ...         "**/*.log",
            ...     ],
            ... )

        Notes:
            - Both `include_patterns` and `exclude_patterns` follow the 'gitignore' pattern format described
              in https://git-scm.com/docs/gitignore#_pattern_format.
            - If both `include_patterns` and `exclude_patterns` are provided, files matching
              `exclude_patterns` will be excluded even if they match `include_patterns`.
        """
        include_spec: typing.Optional[pathspec.PathSpec] = excludespec_from_patterns(
            include_patterns
        )
        exclude_spec: typing.Optional[pathspec.PathSpec] = excludespec_from_patterns(
            exclude_patterns
        )
        all_files = self.__list_directory_files(
            directory_path, include_spec=include_spec, exclude_spec=exclude_spec
        )
        file_destination_paths = {
            path: os.path.relpath(path, directory_path) for path in all_files
        }

        self.upload_files(all_files, file_destination_paths, max_batch_size)

        if delete_after_upload:
            for file in all_files:
                if file.is_file():
                    file.unlink()

    def upload_file(
        self,
        file_path: pathlib.Path,
        file_destination_path: typing.Optional[str] = None,
    ) -> None:
        """
        Upload a single file to the dataset.
        If `file_destination_path` is not provided, the file will be uploaded to the top-level of the dataset.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.upload_file(
            ...     pathlib.Path("/path/to/file.txt"),
            ...     file_destination_path="foo/bar.txt",
            ... )
        """
        if not file_destination_path:
            file_destination_path = file_path.name

        self.upload_files([file_path], {file_path: file_destination_path})

    def upload_files(
        self,
        files: collections.abc.Iterable[pathlib.Path],
        file_destination_paths: collections.abc.Mapping[pathlib.Path, str] = {},
        max_batch_size: int = MAX_FILES_PER_MANIFEST,
    ):
        """
        Upload multiple files to the dataset.
        If `file_destination_paths` is not provided, files will be uploaded to the top-level of the dataset.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.upload_files(
            ...     [
            ...         pathlib.Path("/path/to/file.txt"),
            ...         ...
            ...     ],
            ...     file_destination_paths={
            ...         pathlib.Path("/path/to/file.txt"): "foo/bar.txt",
            ...     },
            ... )
        """
        working_set: list[pathlib.Path] = []

        for file in files:
            working_set.append(file)

            if len(working_set) >= max_batch_size:
                self.__upload_files_batch(working_set, file_destination_paths)
                working_set = []

        if len(working_set) > 0:
            self.__upload_files_batch(working_set, file_destination_paths)

    def _complete_manifest_transaction(self, transaction_id: str) -> None:
        """
        Marks a transaction as 'completed', which allows the Roboto Platform to evaluate triggers
        for automatic action on incoming data. This also aids reporting on partial upload failure cases.

        This should be considered private.

        It is loosely exposed (single underscore instead of double) because of a niche, administrative use-case.
        """
        self.__roboto_client.put(
            f"v2/datasets/{self.dataset_id}/batch_uploads/{transaction_id}/complete"
        )

    def _create_manifest_transaction(
        self,
        origination: str,
        resource_manifest: dict[str, int],
    ) -> tuple[str, dict[str, str]]:
        """
        This should be considered private.

        It is loosely exposed (single underscore instead of double) because of a niche, administrative use-case.
        """
        request = BeginManifestTransactionRequest(
            origination=origination,
            resource_manifest=resource_manifest,
        )

        result = self.__roboto_client.post(
            f"v2/datasets/{self.dataset_id}/batch_uploads",
            data=request,
            caller_org_id=self.org_id,
        ).to_record(BeginManifestTransactionResponse)

        return result.transaction_id, dict(result.upload_mappings)

    def __download_creds_provider(self, bucket_name: str):
        def _wrapped():
            all_creds = self.__get_temporary_credentials(Permissions.ReadOnly)
            filtered = [cred for cred in all_creds if cred.bucket == bucket_name]

            if len(filtered) == 0:
                raise RobotoInternalException(
                    f"Dataset {self.dataset_id} has no read creds for bucket {bucket_name}."
                )

            return filtered[0].to_s3_credentials()

        return _wrapped

    def __upload_creds_provider(self, transaction_id: str):
        def _wrapped():
            all_creds = self.__get_temporary_credentials(
                Permissions.ReadWrite,
                transaction_id=transaction_id,
            )

            if len(all_creds) == 0:
                raise RobotoInternalException(
                    f"Dataset {self.dataset_id} has no write creds."
                )

            return all_creds[0].to_s3_credentials()

        return _wrapped

    def __file_service(self) -> FileService:
        return FileService(self.__roboto_client)

    def __get_temporary_credentials(
        self,
        permissions: Permissions,
        transaction_id: typing.Optional[str] = None,
    ) -> collections.abc.Sequence[DatasetCredentials]:
        query_params = {"mode": permissions.value}

        if transaction_id:
            query_params["transaction_id"] = transaction_id

        return self.__roboto_client.get(
            f"v1/datasets/id/{self.dataset_id}/credentials", query=query_params
        ).to_record_list(DatasetCredentials)

    def _flush_manifest_item_completions(
        self,
        transaction_id: str,
        manifest_items: list[str],
    ) -> None:
        """
        This should be considered private.

        It is loosely exposed (single underscore instead of double) because of a niche, administrative use-case.
        """
        self.__roboto_client.put(
            f"v2/datasets/{self.dataset_id}/batch_uploads/{transaction_id}/progress",
            data=ReportTransactionProgressRequest(
                manifest_items=manifest_items,
            ),
        )

    def __list_directory_files(
        self,
        directory_path: pathlib.Path,
        include_spec: typing.Optional[pathspec.PathSpec] = None,
        exclude_spec: typing.Optional[pathspec.PathSpec] = None,
    ) -> collections.abc.Iterable[pathlib.Path]:
        all_files = set()

        for root, _, files in os.walk(directory_path):
            for file in files:
                should_include = include_spec is None or include_spec.match_file(file)
                should_exclude = exclude_spec is not None and exclude_spec.match_file(
                    file
                )

                if should_include and not should_exclude:
                    all_files.add(pathlib.Path(root, file))

        return all_files

    def __list_files_page(
        self,
        page_token: typing.Optional[str] = None,
        include_patterns: typing.Optional[list[str]] = None,
        exclude_patterns: typing.Optional[list[str]] = None,
    ) -> PaginatedList[FileRecord]:
        """
        List files associated with dataset.

        Files are associated with datasets in an eventually-consistent manner,
        so there will likely be delay between a file being uploaded and it appearing in this list.
        """
        query_params: dict[str, typing.Any] = {}
        if page_token:
            query_params["page_token"] = str(page_token)

        request = QueryDatasetFilesRequest(
            page_token=page_token,
            include_patterns=include_patterns,
            exclude_patterns=exclude_patterns,
        )
        return self.__roboto_client.post(
            f"v1/datasets/{self.dataset_id}/files/query",
            data=request,
            query=query_params,
            idempotent=True,
        ).to_paginated_list(FileRecord)

    def __on_manifest_item_complete(
        self,
        transaction_id: str,
        total_file_count: int,
        manifest_item_identifier: str,
    ) -> None:
        """
        This method is used as a callback (a "subscriber") for S3 TransferManager,
        which is used to upload files to S3.

        TransferManager uses thread-based concurrency under the hood,
        so any instance state accessed or modified here must be synchronized.
        """
        with self.__transaction_completed_mutex:
            if transaction_id not in self.__transaction_completed_unreported_items:
                self.__transaction_completed_unreported_items[transaction_id] = set()

            self.__transaction_completed_unreported_items[transaction_id].add(
                manifest_item_identifier
            )

            completion_count = len(
                self.__transaction_completed_unreported_items[transaction_id]
            )
            if self.__sufficient_uploads_completed_to_report_progress(
                completion_count, total_file_count
            ):
                self._flush_manifest_item_completions(
                    transaction_id=transaction_id,
                    manifest_items=list(
                        self.__transaction_completed_unreported_items[transaction_id]
                    ),
                )
                self.__transaction_completed_unreported_items[transaction_id] = set()

    def __sufficient_uploads_completed_to_report_progress(
        self, completion_count: int, total_file_count: int
    ):
        """
        Determine if there are a sufficient number of files that have already been uploaded
        to S3 to report progress to the Roboto Platform.

        If the total count of files to upload is below the reporting threshold,
        or if for whatever other reason a batch is not large enough to report progress,
        file records will still be finalized as part of the upload finalization routine.
        See, e.g., :py:meth:`~roboto.domain.datasets.dataset.Dataset._complete_manifest_transaction`
        """
        batch_size = math.ceil(total_file_count / Dataset.UPLOAD_REPORTING_BATCH_COUNT)
        return (
            completion_count >= batch_size
            and completion_count >= Dataset.UPLOAD_REPORTING_MIN_BATCH_SIZE
        )

    def __upload_files_batch(
        self,
        files: collections.abc.Iterable[pathlib.Path],
        file_destination_paths: collections.abc.Mapping[pathlib.Path, str] = {},
    ):
        try:
            package_version = importlib.metadata.version("roboto")
        except importlib.metadata.PackageNotFoundError:
            package_version = "version_not_found"

        origination = RobotoEnv.default().roboto_env or f"roboto {package_version}"
        file_manifest = {
            file_destination_paths.get(path, path.name): path.stat().st_size
            for path in files
        }

        total_file_count = len(file_manifest)
        total_file_size = sum(file_manifest.values())

        transaction_id, create_upload_mappings = self._create_manifest_transaction(
            origination=origination,
            resource_manifest=file_manifest,
        )

        file_path_to_manifest_mappings = {
            file_destination_paths.get(path, path.name): path for path in files
        }
        upload_mappings = {
            file_path_to_manifest_mappings[src_path]: dest_uri
            for src_path, dest_uri in create_upload_mappings.items()
        }

        progress_monitor_factory = TqdmProgressMonitorFactory(
            concurrency=1,
            ctx={
                "expected_file_count": total_file_count,
                "expected_file_size": total_file_size,
            },
        )

        with progress_monitor_factory.upload_monitor(
            source=f"{total_file_count} file" + ("s" if total_file_count != 1 else ""),
            size=total_file_size,
        ) as progress_monitor:
            self.__file_service().upload_many_files(
                file_generator=upload_mappings.items(),
                credential_provider=self.__upload_creds_provider(transaction_id),
                on_file_complete=functools.partial(
                    self.__on_manifest_item_complete,
                    transaction_id,
                    total_file_count,
                ),
                progress_monitor=progress_monitor,
                max_concurrency=8,
            )

        self._complete_manifest_transaction(
            transaction_id,
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import random
import typing

import tenacity.wait

RetryWaitFn = typing.Callable[
    [tenacity.RetryCallState, typing.Optional[BaseException]], float
]


def default_retry_wait_ms(
    retry_state: tenacity.RetryCallState, _exc: typing.Optional[BaseException]
) -> float:
    """
    Returns sleep time in ms using exponential backoff with full jitter, as described in:
    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    """
    base = 500
    cap = 30_000

    exponential_wait = min(cap, pow(2, retry_state.attempt_number) * base)
    jittered = random.uniform(0, exponential_wait)
    return jittered



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pydantic


def remove_non_noneable_init_args(data: dict, model: pydantic.BaseModel) -> dict:
    """
    Remove keys from `data` (e.g., kwargs passed to `__init__`) set to `None`
    that are not allowed to be `None` according to `model`'s type definition.

    This is particularly useful when marshalling data from the database into a
    Pydantic model. If a database row has a `NULL` value for a
    field that is not nullable, this function can be used to filter out those `None`
    values before passing the data to the model's `__init__` method.

    Example usage:

    >>> class Foo(pydantic.BaseModel):
    ...     a: int = 3
    ...     def __init__(self, **kwargs):
    ...         filtered_kwargs = remove_non_noneable_init_args(kwargs, self)
    ...         super().__init__(**filtered_kwargs)
    ...
    >>> Foo(a=None)
    Foo(a=3, b='bar')

    Without this function, the above would raise a `ValidationError`.
    """

    def _field_allows_none(field_name):
        field = model.model_fields.get(field_name)
        if field is None:
            return False

        # Allows none if field annotation is a union type and None is one of the types in the union
        try:
            # After dropping 3.9 support, add `isinstance(field.annotation, types.UnionType)`
            # to this statement to remove the `# type: ignore` and ditch the try/except.
            return type(None) in field.annotation.__args__  # type: ignore
        except AttributeError:
            return False

    return {k: v for k, v in data.items() if v is not None or _field_allows_none(k)}


def validate_nonzero_gitpath_specs(value: list[str]) -> list[str]:
    filtered = list(filter(lambda p: p.strip() != "", value))

    if len(value) == 0:
        raise ValueError("Paths must not be empty.")
    elif len(filtered) == 0:
        raise ValueError(
            "Paths must have at least one entry which is not the empty string or just whitespace."
        )

    return value



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import datetime
import typing

from ...exceptions import RobotoConflictException
from ...http import RobotoClient
from ...query import (
    Comparator,
    Condition,
    ConditionType,
    QuerySpecification,
)
from ...sentinels import (
    NotSet,
    NotSetType,
    remove_not_set,
)
from ...waiters import Interval, wait_for
from .action import Action
from .invocation import Invocation
from .invocation_record import (
    ComputeRequirements,
    ContainerParameters,
    InvocationDataSource,
    InvocationSource,
)
from .trigger_operations import (
    CreateTriggerRequest,
    UpdateTriggerRequest,
)
from .trigger_record import (
    TriggerEvaluationRecord,
    TriggerEvaluationStatus,
    TriggerForEachPrimitive,
    TriggerRecord,
)


class Trigger:
    """
    A rule that automatically invokes an action when specific events or conditions occur
    """

    __record: TriggerRecord
    __roboto_client: RobotoClient

    @staticmethod
    def get_evaluations_for_dataset(
        dataset_id: str,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator[TriggerEvaluationRecord, None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        page_token: typing.Optional[str] = None
        while True:
            query_params: dict[str, typing.Any] = {}
            if page_token is not None:
                query_params["page_token"] = page_token

            paginated_results = roboto_client.get(
                f"v1/triggers/dataset/id/{dataset_id}/evaluations",
                query=query_params,
                owner_org_id=owner_org_id,
            ).to_paginated_list(TriggerEvaluationRecord)
            for record in paginated_results.items:
                yield record
            if paginated_results.next_token:
                page_token = paginated_results.next_token
            else:
                break

    @classmethod
    def create(
        cls,
        name: str,
        action_name: str,
        required_inputs: list[str],
        for_each: TriggerForEachPrimitive,
        enabled: bool = True,
        action_digest: typing.Optional[str] = None,
        action_owner_id: typing.Optional[str] = None,
        additional_inputs: typing.Optional[list[str]] = None,
        compute_requirement_overrides: typing.Optional[ComputeRequirements] = None,
        condition: typing.Optional[ConditionType] = None,
        container_parameter_overrides: typing.Optional[ContainerParameters] = None,
        parameter_values: typing.Optional[dict[str, typing.Any]] = None,
        service_user_id: typing.Optional[str] = None,
        timeout: typing.Optional[int] = None,
        caller_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Trigger":
        """
        Invoke an action on every new dataset (or every new dataset file) that meets some acceptance criteria.
        """
        roboto_client = RobotoClient.defaulted(roboto_client)

        request = CreateTriggerRequest(
            action_digest=action_digest,
            action_name=action_name,
            action_owner_id=action_owner_id,
            additional_inputs=additional_inputs,
            compute_requirement_overrides=compute_requirement_overrides,
            condition=condition,
            container_parameter_overrides=container_parameter_overrides,
            enabled=enabled,
            for_each=for_each,
            name=name,
            parameter_values=parameter_values,
            required_inputs=required_inputs,
            service_user_id=service_user_id,
            timeout=timeout,
        )

        response = roboto_client.post(
            "v1/triggers",
            data=request,
            caller_org_id=caller_org_id,
        )
        record = response.to_record(TriggerRecord)
        return cls(record, roboto_client)

    @classmethod
    def from_name(
        cls,
        name: str,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Trigger":
        roboto_client = RobotoClient.defaulted(roboto_client)
        response = roboto_client.get(
            f"v1/triggers/{name}",
            owner_org_id=owner_org_id,
        )
        record = response.to_record(TriggerRecord)
        return cls(record, roboto_client)

    @classmethod
    def query(
        cls,
        spec: typing.Optional[QuerySpecification] = None,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator["Trigger", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        spec = spec or QuerySpecification()

        while True:
            response = roboto_client.post(
                "v1/triggers/query",
                data=spec,
                owner_org_id=owner_org_id,
                idempotent=True,
            )
            paginated_results = response.to_paginated_list(TriggerRecord)
            for record in paginated_results.items:
                yield cls(record, roboto_client)
            if paginated_results.next_token:
                spec.after = paginated_results.next_token
            else:
                break

    def __init__(
        self,
        record: TriggerRecord,
        roboto_client: typing.Optional[RobotoClient] = None,
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def condition(self) -> typing.Optional[ConditionType]:
        return self.__record.condition

    @property
    def created(self) -> datetime.datetime:
        return self.__record.created

    @property
    def created_by(self) -> str:
        return self.__record.created_by

    @property
    def enabled(self) -> bool:
        return self.__record.enabled

    @property
    def for_each(self) -> TriggerForEachPrimitive:
        return self.__record.for_each

    @property
    def name(self):
        return self.__record.name

    @property
    def modified(self) -> datetime.datetime:
        return self.__record.modified

    @property
    def modified_by(self) -> str:
        return self.__record.modified_by

    @property
    def org_id(self):
        return self.__record.org_id

    @property
    def record(self) -> TriggerRecord:
        return self.__record

    @property
    def service_user_id(self) -> typing.Optional[str]:
        return self.__record.service_user_id

    @property
    def trigger_id(self) -> str:
        return self.__record.trigger_id

    def delete(self):
        self.__roboto_client.delete(
            f"v1/triggers/{self.name}",
            owner_org_id=self.org_id,
        )

    def disable(self):
        self.update(enabled=False)

    def enable(self):
        self.update(enabled=True)

    def get_action(self) -> Action:
        return Action.from_name(
            self.__record.action.name,
            digest=self.__record.action.digest,
            owner_org_id=self.__record.action.owner,
            roboto_client=self.__roboto_client,
        )

    def get_evaluations(
        self,
        limit: typing.Optional[int] = None,
        page_token: typing.Optional[str] = None,
    ) -> collections.abc.Generator[TriggerEvaluationRecord, None, None]:
        query_params: dict[str, typing.Union[int, str]] = {}
        if limit is not None:
            query_params["limit"] = limit

        while True:
            if page_token is not None:
                query_params["page_token"] = page_token
            response = self.__roboto_client.get(
                f"v1/triggers/{self.name}/evaluations",
                query=query_params,
                owner_org_id=self.org_id,
            )
            paginated_results = response.to_paginated_list(TriggerEvaluationRecord)
            for record in paginated_results.items:
                yield record
            if paginated_results.next_token:
                page_token = paginated_results.next_token
            else:
                break

    def get_invocations(self) -> collections.abc.Generator[Invocation, None, None]:
        query_spec = QuerySpecification(
            condition=Condition(
                field="provenance.source.source_id",
                comparator=Comparator.Equals,
                value=self.trigger_id,
            )
        )
        yield from Invocation.query(
            query_spec,
            owner_org_id=self.org_id,
            roboto_client=self.__roboto_client,
        )

    def latest_evaluation(self) -> typing.Optional[TriggerEvaluationRecord]:
        evaluations = list(self.get_evaluations(limit=1))
        return evaluations[0] if evaluations else None

    def invoke(
        self,
        data_source: InvocationDataSource,
        idempotency_id: typing.Optional[str] = None,
        input_data_override: typing.Optional[list[str]] = None,
    ) -> typing.Optional[Invocation]:
        try:
            return self.get_action().invoke(
                compute_requirement_overrides=self.__record.compute_requirement_overrides,
                container_parameter_overrides=self.__record.container_parameter_overrides,
                data_source_id=data_source.data_source_id,
                data_source_type=data_source.data_source_type,
                input_data=input_data_override or self.__record.required_inputs,
                idempotency_id=idempotency_id,
                invocation_source=InvocationSource.Trigger,
                invocation_source_id=self.__record.name,
                parameter_values=self.__record.parameter_values,
                timeout=self.__record.timeout,
                caller_org_id=self.__record.org_id,
            )
        except RobotoConflictException:
            # Return None if there was an existing invocation with the same idempotency ID
            return None

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")

    def update(
        self,
        action_name: typing.Union[str, NotSetType] = NotSet,
        action_owner_id: typing.Union[str, NotSetType] = NotSet,
        action_digest: typing.Optional[typing.Union[str, NotSetType]] = NotSet,
        additional_inputs: typing.Optional[
            typing.Union[list[str], NotSetType]
        ] = NotSet,
        compute_requirement_overrides: typing.Optional[
            typing.Union[ComputeRequirements, NotSetType]
        ] = NotSet,
        container_parameter_overrides: typing.Optional[
            typing.Union[ContainerParameters, NotSetType]
        ] = NotSet,
        condition: typing.Optional[typing.Union[ConditionType, NotSetType]] = NotSet,
        enabled: typing.Union[bool, NotSetType] = NotSet,
        for_each: typing.Union[TriggerForEachPrimitive, NotSetType] = NotSet,
        parameter_values: typing.Optional[
            typing.Union[dict[str, typing.Any], NotSetType]
        ] = NotSet,
        required_inputs: typing.Union[list[str], NotSetType] = NotSet,
        timeout: typing.Optional[typing.Union[int, NotSetType]] = NotSet,
    ) -> "Trigger":
        request = remove_not_set(
            UpdateTriggerRequest(
                action_name=action_name,
                action_owner_id=action_owner_id,
                action_digest=action_digest,
                additional_inputs=additional_inputs,
                compute_requirement_overrides=compute_requirement_overrides,
                container_parameter_overrides=container_parameter_overrides,
                condition=condition,
                enabled=enabled,
                for_each=for_each,
                parameter_values=parameter_values,
                required_inputs=required_inputs,
                timeout=timeout,
            )
        )

        response = self.__roboto_client.put(
            f"v1/triggers/{self.name}",
            data=request,
            owner_org_id=self.org_id,
        )
        record = response.to_record(TriggerRecord)
        self.__record = record
        return self

    def wait_for_evaluations_to_complete(
        self,
        timeout: float = 60 * 5,  # 5 minutes in seconds
        poll_interval: Interval = 5,
    ) -> None:
        """
        Wait for all evaluations for this trigger to complete.

        Throws a :py:exc:`~roboto.waiters.TimeoutError` if the timeout is reached.

        Args:
            timeout: The maximum amount of time, in seconds, to wait for the evaluations to complete.
            poll_interval: The amount of time, in seconds, to wait between polling iterations.
        """
        wait_for(
            lambda: all(
                [
                    evaluation.status != TriggerEvaluationStatus.Pending
                    for evaluation in self.get_evaluations()
                ]
            ),
            timeout=timeout,
            interval=poll_interval,
            timeout_msg=f"Timed out waiting for evaluations for trigger '{self.name}' to complete",
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from roboto.sentinels import (
    NotSet,
    value_or_not_set,
)

from ...domain.collections import (
    Collection,
    CollectionResourceRef,
    CollectionResourceType,
)
from ..command import RobotoCommand
from ..context import CLIContext
from .shared_helpdoc import COLLECTION_ID_HELP


def update(args, context: CLIContext, parser: argparse.ArgumentParser):
    collection = Collection.from_id(
        collection_id=args.collection_id,
        roboto_client=context.roboto_client,
    )

    add_resources: list[CollectionResourceRef] = []
    if args.add_dataset_id:
        add_resources.extend(
            [
                CollectionResourceRef(
                    resource_type=CollectionResourceType.Dataset, resource_id=dataset_id
                )
                for dataset_id in args.add_dataset_id
            ]
        )

    if args.add_file_id:
        add_resources.extend(
            [
                CollectionResourceRef(
                    resource_type=CollectionResourceType.File, resource_id=file_id
                )
                for file_id in args.add_file_id
            ]
        )

    remove_resources: list[CollectionResourceRef] = []
    if args.remove_dataset_id:
        remove_resources.extend(
            [
                CollectionResourceRef(
                    resource_type=CollectionResourceType.Dataset, resource_id=dataset_id
                )
                for dataset_id in args.remove_dataset_id
            ]
        )

    if args.remove_file_id:
        remove_resources.extend(
            [
                CollectionResourceRef(
                    resource_type=CollectionResourceType.File, resource_id=file_id
                )
                for file_id in args.remove_file_id
            ]
        )

    collection.update(
        name=value_or_not_set(args.name),
        description=value_or_not_set(args.description),
        add_tags=value_or_not_set(args.add_tag),
        remove_tags=value_or_not_set(args.remove_tag),
        add_resources=NotSet if len(add_resources) == 0 else add_resources,
        remove_resources=NotSet if len(remove_resources) == 0 else remove_resources,
    )

    print(collection.record.model_dump_json(indent=2))


def update_setup_parser(parser):
    parser.add_argument("collection_id", type=str, help=COLLECTION_ID_HELP)
    parser.add_argument(
        "--name",
        type=str,
        help="A human-readable name for this collection. "
        + "Does not need to be unique.",
    )
    parser.add_argument(
        "--description", type=str, help="Information about what's in this collection"
    )
    parser.add_argument(
        "--add-dataset-id",
        nargs="*",
        help="Datasets to add to this collection.",
        action="extend",
    )
    parser.add_argument(
        "--add-file-id",
        nargs="*",
        help="Files to add to this collection.",
        action="extend",
    )
    parser.add_argument(
        "--remove-dataset-id",
        nargs="*",
        help="Datasets to remove from this collection.",
        action="extend",
    )
    parser.add_argument(
        "--remove-file-id",
        nargs="*",
        help="Files to remove from this collection.",
        action="extend",
    )
    parser.add_argument(
        "--add-tag",
        type=str,
        nargs="*",
        help="Tags to add to this collection.",
        action="extend",
    )
    parser.add_argument(
        "--remove-tag",
        type=str,
        nargs="*",
        help="Tags to remove from this collection.",
        action="extend",
    )


update_command = RobotoCommand(
    name="update",
    logic=update,
    setup_parser=update_setup_parser,
    command_kwargs={
        "help": "Updates the resources and metadata of a given collection."
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...domain.tokens import Token
from ..command import (
    RobotoCommand,
    RobotoCommandSet,
)
from ..context import CLIContext


def create(args, context: CLIContext, parser: argparse.ArgumentParser):
    token, secret = Token.create(
        expiry_days=args.expiry_days,
        name=args.name,
        description=args.description,
        roboto_client=context.roboto_client,
    )

    creds_example_json = json.dumps(
        {"username": "<your email>", "token": secret}, indent=2
    )

    print(
        "This secret will only be available to you once, so store it somewhere safe!\n"
        + "To use this personal access token, create a file at ~/.roboto/config.json, and populate it with "
        + f"{creds_example_json}."
    )


def create_setup_parser(parser):
    parser.add_argument(
        "--name",
        type=str,
        required=True,
        help="A human readable name for this token.",
    )

    parser.add_argument(
        "--description",
        type=str,
        help="An optional description for this token.",
    )

    parser.add_argument(
        "--expiry-days",
        type=int,
        choices=[30, 60, 90],
        default=30,
        help="The number of days this token will be valid for.",
    )


def list(args, context: CLIContext, parser: argparse.ArgumentParser):
    tokens = Token.for_self(roboto_client=context.roboto_client)
    for token in tokens:
        print(json.dumps(token.to_dict(), indent=2))


def show(args, context: CLIContext, parser: argparse.ArgumentParser):
    token = Token.from_id(token_id=args.id, roboto_client=context.roboto_client)
    print(json.dumps(token.to_dict(), indent=2))


def show_setup_parser(parser):
    parser.add_argument(
        "--id",
        type=str,
        required=True,
        help="The token_id for a token to look up.",
    )


def delete(args, context: CLIContext, parser: argparse.ArgumentParser):
    Token.from_id(token_id=args.id, roboto_client=context.roboto_client).delete()
    print(f"Successfully deleted token '{args.id}'!")


def delete_setup_parser(parser):
    parser.add_argument(
        "--id",
        type=str,
        required=True,
        help="The token_id for a token to delete.",
    )


create_command = RobotoCommand(
    name="create",
    logic=create,
    setup_parser=create_setup_parser,
    command_kwargs={"help": "Creates a temporary access token."},
)


list_command = RobotoCommand(
    name="list",
    logic=list,
    command_kwargs={"help": "Creates a temporary access token."},
)

show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_setup_parser,
    command_kwargs={"help": "Gets a token by unique token_id."},
)

delete_command = RobotoCommand(
    name="delete",
    logic=delete,
    setup_parser=delete_setup_parser,
    command_kwargs={"help": "Deletes a token by unique token_id."},
)

commands = [
    create_command,
    delete_command,
    list_command,
    show_command,
]

command_set = RobotoCommandSet(
    name="tokens",
    help="Generate and review your platform access tokens.",
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic
from pydantic import ConfigDict

from ...pydantic import (
    remove_non_noneable_init_args,
)
from ...updates import (
    MetadataChangeset,
    UpdateCondition,
)


class BeginSingleFileUploadRequest(pydantic.BaseModel):
    """
    Request payload to begin a single file upload
    """

    origination: typing.Optional[str] = pydantic.Field(
        description="Additional information about what uploaded the file, e.g. `roboto client v1.0.0`.",
        default=None,
    )
    file_path: str = pydantic.Field(
        description="The destination path to upload the file to in the dataset, e.g. `recording.bag`"
        + "or `path/to/metadata.json`.",
    )
    file_size: int = pydantic.Field(description="The size of the file in bytes.")


class BeginSingleFileUploadResponse(pydantic.BaseModel):
    """
    Response to a single file upload
    """

    upload_id: str
    upload_url: str


class BeginManifestTransactionRequest(pydantic.BaseModel):
    """
    Request payload to begin a manifest-based transaction
    """

    origination: str
    resource_manifest: dict[str, int]


class BeginManifestTransactionResponse(pydantic.BaseModel):
    """
    Response to a manifest-based transaction request
    """

    transaction_id: str
    upload_mappings: dict[str, str]


class CreateDatasetRequest(pydantic.BaseModel):
    """
    Request payload to create a dataset
    """

    description: typing.Optional[str] = pydantic.Field(
        default=None,
        description="An optional human-readable description for this dataset.",
    )
    metadata: dict[str, typing.Any] = pydantic.Field(
        default_factory=dict,
        description="Initial key-value pairs to associate with this dataset for discovery and search, e.g. "
        + "`{ 'softwareVersion': '3.1.4' }`",
    )
    tags: list[str] = pydantic.Field(
        default_factory=list,
        description="Initial tags to associate with this dataset for discovery and search, e.g. "
        + "`['sunny', 'campaign5']`",
    )

    def __init__(self, **data):
        super().__init__(**remove_non_noneable_init_args(data, self))


class QueryDatasetFilesRequest(pydantic.BaseModel):
    """
    Request payload to query files in a dataset
    """

    page_token: typing.Optional[str] = None
    include_patterns: typing.Optional[list[str]] = None
    exclude_patterns: typing.Optional[list[str]] = None


class QueryDatasetsRequest(pydantic.BaseModel):
    """
    Request payload to query datasets
    """

    filters: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    model_config = ConfigDict(extra="forbid")


class ReportTransactionProgressRequest(pydantic.BaseModel):
    """
    Request payload to report transaction progress
    """

    manifest_items: list[str]


class UpdateDatasetRequest(pydantic.BaseModel):
    """
    Request payload to update a dataset
    """

    metadata_changeset: typing.Optional[MetadataChangeset] = None
    description: typing.Optional[str] = None
    conditions: typing.Optional[list[UpdateCondition]] = None


class BeginTransactionRequest(pydantic.BaseModel):
    """
    Request payload to begin a transaction
    """

    origination: str
    expected_resource_count: typing.Optional[int] = None


class TransactionCompletionResponse(pydantic.BaseModel):
    """
    Response to transaction completion
    """

    is_complete: bool


class DeleteDirectoriesRequest(pydantic.BaseModel):
    """
    Request payload to delete directories in a dataset
    """

    directory_paths: list[str]


class RenameDirectoryRequest(pydantic.BaseModel):
    """
    Request payload to rename a directory in a dataset
    """

    new_path: str
    old_path: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import time
from typing import Any, Optional

from ...domain import actions
from ..command import RobotoCommand
from ..context import CLIContext


def get_logs(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    invocation = actions.Invocation.from_id(
        args.invocation_id,
        roboto_client=context.roboto_client,
    )

    process_decorator = "=" * 10
    last_log = None
    for log_record in invocation.get_logs():
        if last_log is None:
            print(f"{process_decorator} {log_record.process.value} {process_decorator}")
        else:
            if (
                last_log.partial_id is not None
                and log_record.partial_id != last_log.partial_id
            ):
                # `last_record` was the last partial message from its group--print it
                print(last_log.log)

            if last_log.process != log_record.process:
                print()  # Newline separate process output
                print(
                    f"{process_decorator} {log_record.process.value} {process_decorator}"
                )

        if log_record.partial_id is None:
            print(log_record.log)

        last_log = log_record


def stream_logs(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    invocation = actions.Invocation.from_id(
        args.invocation_id,
        roboto_client=context.roboto_client,
    )

    last_read: Optional[Any] = None
    wait_msg = ""
    last_log = None
    process_decorator = "=" * 10
    try:
        while True:
            try:
                log_record_generator = invocation.stream_logs(last_read)
                while True:
                    log_record = next(log_record_generator)
                    if wait_msg:
                        # Clear wait message
                        print("\r", " " * len(wait_msg), end="\r", flush=True)
                        wait_msg = ""

                    if last_log is None:
                        print(
                            f"{process_decorator} {log_record.process.value} {process_decorator}"
                        )
                    else:
                        if (
                            last_log.partial_id is not None
                            and log_record.partial_id != last_log.partial_id
                        ):
                            # `last_record` was the last partial message from its group--print it
                            print(last_log.log)

                        if last_log.process != log_record.process:
                            print()  # Newline separate process output
                            print(
                                f"{process_decorator} {log_record.process.value} {process_decorator}"
                            )

                    if log_record.partial_id is None:
                        print(log_record.log)

                    last_log = log_record
            except StopIteration as stop:
                if invocation.reached_terminal_status:
                    break

                if not wait_msg:
                    wait_msg = "Waiting for logs..."
                    print(wait_msg, end="", flush=True)

                last_read = stop.value
                time.sleep(2)
                invocation.refresh()
    except KeyboardInterrupt:
        pass  # Swallow

    if wait_msg:
        print("\r", " " * len(wait_msg), end="\r", flush=True)


def main(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    if args.tail:
        stream_logs(args, context, parser)
    else:
        get_logs(args, context, parser)


def get_logs_parser(parser: argparse.ArgumentParser):
    parser.add_argument("invocation_id")
    parser.add_argument("--tail", required=False, action="store_true")


get_logs_command = RobotoCommand(
    name="logs",
    logic=main,
    setup_parser=get_logs_parser,
    command_kwargs={"help": "Get invocation logs."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import http
import json
import typing
import urllib.error


class HttpError(Exception):
    __http_exc: urllib.error.HTTPError
    __msg: typing.Any = None
    __status: typing.Optional[http.HTTPStatus] = None

    def __init__(self, exc: urllib.error.HTTPError) -> None:
        super().__init__()
        self.__http_exc = exc

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.msg!r}, {self.status!r})"

    def __str__(self) -> str:
        return f"{self.msg!r}"

    @property
    def status(self) -> typing.Optional[http.HTTPStatus]:
        if self.__status is None:
            try:
                self.__status = http.HTTPStatus(int(self.__http_exc.code))
            except ValueError:
                self.__status = None
        return self.__status

    @property
    def msg(self) -> typing.Any:
        if self.__msg is None:
            decoded = self.__http_exc.read().decode("utf-8")
            try:
                self.__msg = json.loads(decoded)
            except json.JSONDecodeError:
                self.__msg = decoded
        return self.__msg

    @property
    def headers(self) -> dict:
        return dict(self.__http_exc.headers.items())


class ClientError(HttpError):
    pass


class ServerError(HttpError):
    pass



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json
import typing

from ...domain import actions
from ...sentinels import NotSet, is_set, null
from ...updates import MetadataChangeset
from ..command import (
    KeyValuePairsAction,
    RobotoCommand,
)
from ..common_args import (
    ActionParameterArg,
    ActionTimeoutArg,
    add_compute_requirements_args,
    add_container_parameters_args,
    add_org_arg,
    parse_compute_requirements,
    parse_container_overrides,
)
from ..context import CLIContext
from ..terminal import print_error_and_exit
from ..validation import (
    pydantic_validation_handler,
)


def update(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    action = actions.Action.from_name(
        name=args.action,
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    )

    updates: dict[str, typing.Any] = dict()

    if args.description is null:
        updates["description"] = None
    elif args.description != NotSet:
        updates["description"] = args.description if args.description else None

    if args.short_description is null:
        updates["short_description"] = None
    elif args.short_description != NotSet:
        updates["short_description"] = (
            args.short_description if args.short_description else None
        )

    if is_set(args.timeout):
        updates["timeout"] = args.timeout

    if args.image != NotSet:
        stripped = args.image.strip()
        if not stripped:
            raise ValueError(
                "image can neither be unset nor set to an empty (or whitespace-only) string"
            )
        updates["uri"] = stripped
    with pydantic_validation_handler("Action parameter updates"):
        parameter_changeset = actions.ActionParameterChangeset(
            put_parameters=args.put_parameters if args.put_parameters else [],
            remove_parameters=args.remove_parameters if args.remove_parameters else [],
        )
        if not parameter_changeset.is_empty():
            updates["parameter_changeset"] = parameter_changeset

    with pydantic_validation_handler("metadata updates"):
        metadata_changeset = MetadataChangeset(
            put_tags=args.put_tags,
            remove_tags=args.remove_tags,
            put_fields=args.put_metadata,
            remove_fields=args.remove_metadata,
        )
        if not metadata_changeset.is_empty():
            updates["metadata_changeset"] = metadata_changeset

    compute_requirements = parse_compute_requirements(args, action.compute_requirements)
    if compute_requirements:
        updates["compute_requirements"] = compute_requirements

    container_parameters = parse_container_overrides(args, action.container_parameters)
    if container_parameters:
        updates["container_parameters"] = container_parameters

    if not updates:
        print_error_and_exit("No updates specified. Nothing to do.")

    action.update(**updates)

    print(f"Successfully updated action '{action.name}'. Record: ")
    print(json.dumps(action.to_dict(), indent=2))


def update_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "action",
        metavar="action_reference: <action_name>",
        help="Exact name of action to update.",
    )
    parser.add_argument(
        "--description",
        required=False,
        action="store",
        type=lambda s: s if s != "null" else null,
        default=NotSet,
        help="Optional description of action. Specify ``null`` to unset existing description.",
    )
    parser.add_argument(
        "--short-description",
        required=False,
        action="store",
        type=lambda s: s if s != "null" else null,
        default=NotSet,
        help="Optional short description of an action. Specify ``null`` to unset existing description.",
    )
    parser.add_argument(
        "--image",
        required=False,
        action="store",
        dest="image",
        type=str,
        default=NotSet,
        help="Register a Docker image with this action. Can neither be unset nor set an empty string.",
    )
    parser.add_argument(
        "--timeout",
        required=False,
        action=ActionTimeoutArg,
        type=lambda s: s if s != "null" else None,
        default=NotSet,
        help="Optional timeout for an action in minutes, defaults to 30 minutes or 12 hours depending on tier.",
    )
    parser.add_argument(
        "--put-parameter",
        dest="put_parameters",
        required=False,
        metavar=ActionParameterArg.METAVAR,
        nargs="*",
        action=ActionParameterArg,
        help=(
            "Add parameter(s) or overwrite existing parameter(s) with the same name. "
            "Argument values must be wrapped in quotes. E.g.: "
            "``--put-parameter 'name=my_param|required=true|description=My description of my_param'``"
        ),
    )
    parser.add_argument(
        "--remove-parameter",
        dest="remove_parameters",
        required=False,
        metavar="PARAMETER_NAME",
        nargs="*",
        help="Remove parameter(s) with the given name(s).",
    )
    parser.add_argument(
        "--put-tags",
        help="Add each tag in this sequence if it doesn't exist",
        nargs="*",  # 0 or more
    )

    parser.add_argument(
        "--remove-tags",
        help="Remove each tag in this sequence if it exists",
        nargs="*",  # 0 or more
    )

    parser.add_argument(
        "--put-metadata",
        required=False,
        metavar="KEY_PATH=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``key_path=value`` formatted pairs. "
            "An attempt is made to parse ``value`` as JSON; if this fails, ``value`` is stored as a string. "
            "If ``key_path`` already exists, existing value will be overwritten. "
            "Dot notation is supported for nested keys. "
            "Examples: "
            "``--put-metadata 'key1=value1' 'key2.subkey1=value2' 'key3.sublist1=[\"a\",\"b\",\"c\"]'``"  # noqa: E501
        ),
    )

    parser.add_argument(
        "--remove-metadata",
        required=False,
        metavar="KEY_PATH",
        nargs="*",
        help=(
            "Remove each key from dataset metadata if it exists. "
            "Dot notation is supported for nested keys. E.g.: ``--remove-metadata key1 key2.subkey3``"
        ),
    )
    add_org_arg(parser=parser)
    add_compute_requirements_args(parser)
    add_container_parameters_args(parser)


update_command = RobotoCommand(
    name="update",
    logic=update,
    setup_parser=update_parser,
    command_kwargs={"help": "Update an existing action."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import functools
import typing


def get_by_path(
    target: collections.abc.Mapping[typing.Any, typing.Any],
    key_path: collections.abc.Iterable[typing.Any],
) -> typing.Any:
    """
    Access a key path in a mapping.
    Returns `None` if any part of the path is not found or traverses through an object that is not a mapping.
    """
    try:
        return functools.reduce(lambda d, key: d.get(key, None), key_path, target)
    except (TypeError, AttributeError):
        return None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...domain.datasets import Dataset
from ...updates import MetadataChangeset
from ..command import (
    KeyValuePairsAction,
    RobotoCommand,
)
from ..context import CLIContext
from .shared_helpdoc import DATASET_ID_HELP


def update(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    metadata_changeset = MetadataChangeset(
        put_tags=args.put_tags,
        remove_tags=args.remove_tags,
        put_fields=args.put_metadata,
        remove_fields=args.remove_metadata,
    )
    if metadata_changeset.is_empty() and not args.description:
        parser.error("No dataset changes specified.")

    dataset = Dataset.from_id(args.dataset_id, context.roboto_client)
    dataset.update(metadata_changeset=metadata_changeset, description=args.description)

    print(f"Successfully updated dataset '{dataset.dataset_id}'. Record: ")
    print(json.dumps(dataset.to_dict(), indent=2))


def update_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-d", "--dataset-id", type=str, required=True, help=DATASET_ID_HELP
    )

    parser.add_argument(
        "--description", help="A new description to add to this dataset"
    )

    parser.add_argument(
        "--put-tags",
        help="Add each tag in this sequence if it doesn't exist",
        nargs="*",  # 0 or more
    )

    parser.add_argument(
        "--remove-tags",
        help="Remove each tag in this sequence if it exists",
        nargs="*",  # 0 or more
    )

    parser.add_argument(
        "--put-metadata",
        required=False,
        metavar="KEY_PATH=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``<key>=<value>`` formatted pairs. "
            "An attempt is made to parse ``value`` as JSON; if this fails, ``value`` is stored as a string. "
            "If ``key`` already exists, existing value will be overwritten. "
            "Dot notation is supported for nested keys. "
            "Examples: "
            "``--put-metadata 'key1=value1' 'key2.subkey1=value2' 'key3.sublist1=[\"a\",\"b\",\"c\"]'``"  # noqa: E501
        ),
    )

    parser.add_argument(
        "--remove-metadata",
        required=False,
        metavar="KEY_PATH",
        nargs="*",
        help=(
            "Remove each key from dataset metadata if it exists. "
            "Dot notation is supported for nested keys. E.g.: ``--remove-metadata key1 key2.subkey3``"
        ),
    )


update_command = RobotoCommand(
    name="update",
    logic=update,
    setup_parser=update_parser,
    command_kwargs={"help": "Update an existing dataset."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
from functools import partial
import pathlib
from typing import Any, Callable, Optional
import urllib.parse

import boto3.s3.transfer as s3_transfer

from ...http import RobotoClient
from ...logging import default_logger
from .file import File
from .progress import (
    NoopProgressMonitor,
    NoopProgressMonitorFactory,
    ProgressMonitor,
    ProgressMonitorFactory,
)
from .record import (
    CredentialProvider,
    FileRecord,
    FileTag,
)

# Used to change between showing progress bars for every file and "uploading X files"
MANY_FILES = 100


logger = default_logger()


class DynamicCallbackSubscriber(s3_transfer.BaseSubscriber):
    __on_done_cb: Optional[Callable[[], None]]
    __on_progress_cb: Optional[Callable[[Any], None]]
    __on_queued_cb: Optional[Callable[[Any], None]]

    def __init__(
        self,
        on_done_cb: Optional[Callable[[], None]] = None,
        on_progress_cb: Optional[Callable[[Any], None]] = None,
        on_queued_cb: Optional[Callable[[Any], None]] = None,
    ):
        self.__on_done_cb = on_done_cb
        self.__on_progress_cb = on_progress_cb
        self.__on_queued_cb = on_queued_cb

    def on_queued(self, future, **kwargs):
        if self.__on_queued_cb is not None:
            self.__on_queued_cb(future)

    def on_progress(self, future, bytes_transferred, **kwargs):
        if self.__on_progress_cb is not None:
            self.__on_progress_cb(future)

    def on_done(self, future, **kwargs):
        if self.__on_done_cb is not None:
            self.__on_done_cb()


class FileService:
    __roboto_client: RobotoClient

    @staticmethod
    def __transfer_manager_for_client_provider(
        credential_provider: CredentialProvider, max_concurrency: int = 8
    ):
        s3_client = File.generate_s3_client(credential_provider)
        transfer_config = s3_transfer.TransferConfig(
            use_threads=True, max_concurrency=max_concurrency
        )
        return s3_transfer.create_transfer_manager(s3_client, transfer_config)

    def __init__(self, roboto_client: RobotoClient):
        self.__roboto_client = roboto_client

    def download_files(
        self,
        file_generator: collections.abc.Generator[
            tuple[FileRecord, pathlib.Path], None, None
        ],
        credential_provider: CredentialProvider,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
        max_concurrency: int = 8,
    ) -> None:
        total_file_count = progress_monitor_factory.get_context().get(
            "total_file_count", 0
        )

        if total_file_count >= 20:
            self.__download_many_files(
                file_generator,
                credential_provider,
                progress_monitor_factory,
                max_concurrency,
            )
        else:
            for record, local_path in file_generator:
                File(record, self.__roboto_client).download(
                    local_path, credential_provider, progress_monitor_factory
                )

    def upload_files(
        self,
        bucket: str,
        file_generator: collections.abc.Generator[tuple[pathlib.Path, str], None, None],
        credential_provider: CredentialProvider,
        tags: Optional[dict[FileTag, str]] = None,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
        max_concurrency: int = 20,
    ) -> None:
        expected_file_count = progress_monitor_factory.get_context().get(
            "expected_file_count", "?"
        )

        if expected_file_count >= MANY_FILES:
            base_path = progress_monitor_factory.get_context().get("base_path", "?")
            expected_file_size = progress_monitor_factory.get_context().get(
                "expected_file_size", -1
            )

            progress_monitor = progress_monitor_factory.upload_monitor(
                source=f"{expected_file_count} files from {base_path}",
                size=expected_file_size,
            )
            self.upload_many_files(
                file_generator=file_generator,
                credential_provider=credential_provider,
                max_concurrency=max_concurrency,
                progress_monitor=progress_monitor,
                tags=tags,
            )
        else:
            for src, key in file_generator:
                self.upload_file(
                    local_path=src,
                    bucket=bucket,
                    key=key,
                    credential_provider=credential_provider,
                    tags=tags,
                    progress_monitor_factory=progress_monitor_factory,
                )

    def upload_many_files(
        self,
        file_generator: collections.abc.Iterable[tuple[pathlib.Path, str]],
        credential_provider: CredentialProvider,
        on_file_complete: Optional[Callable[[str, str], None]] = None,
        tags: Optional[dict[FileTag, str]] = None,
        progress_monitor: ProgressMonitor = NoopProgressMonitor(),
        max_concurrency: int = 20,
    ):
        extra_args: Optional[dict[str, Any]] = None
        if tags is not None:
            serializable_tags = {tag.value: value for tag, value in tags.items()}
            encoded_tags = urllib.parse.urlencode(serializable_tags)
            extra_args = {"Tagging": encoded_tags}

        with self.__transfer_manager_for_client_provider(
            credential_provider, max_concurrency
        ) as transfer_manager:
            for src, uri in file_generator:
                parsed_uri = urllib.parse.urlparse(uri)
                bucket = parsed_uri.netloc
                key = parsed_uri.path.lstrip("/")

                subscribers = [
                    s3_transfer.ProgressCallbackInvoker(progress_monitor.update)
                ]

                if on_file_complete is not None:
                    subscribers.append(
                        DynamicCallbackSubscriber(
                            on_done_cb=partial(on_file_complete, uri)
                        )
                    )

                transfer_manager.upload(
                    str(src),
                    bucket,
                    key,
                    extra_args=extra_args,
                    subscribers=subscribers,
                )

    def upload_file(
        self,
        local_path: pathlib.Path,
        bucket: str,
        key: str,
        credential_provider: CredentialProvider,
        tags: Optional[dict[FileTag, str]] = None,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
    ) -> None:
        upload_file_args: dict[str, Any] = {
            "Filename": str(local_path),
            "Key": key,
            "Bucket": bucket,
        }

        if tags is not None:
            serializable_tags = {tag.value: value for tag, value in tags.items()}
            encoded_tags = urllib.parse.urlencode(serializable_tags)
            upload_file_args["ExtraArgs"] = {"Tagging": encoded_tags}

        progress_monitor = progress_monitor_factory.upload_monitor(
            source=key, size=local_path.stat().st_size
        )
        upload_file_args["Callback"] = progress_monitor.update

        try:
            s3_client = File.generate_s3_client(credential_provider)
            s3_client.upload_file(**upload_file_args)
        finally:
            if progress_monitor is not None:
                progress_monitor.close()

    def __download_many_files(
        self,
        file_generator: collections.abc.Generator[
            tuple[FileRecord, pathlib.Path], None, None
        ],
        credential_provider: CredentialProvider,
        progress_monitor_factory: ProgressMonitorFactory = NoopProgressMonitorFactory(),
        max_concurrency: int = 8,
    ):
        transfer_manager = self.__transfer_manager_for_client_provider(
            credential_provider, max_concurrency
        )

        total_file_count = progress_monitor_factory.get_context().get(
            "total_file_count", 0
        )

        base_path = progress_monitor_factory.get_context().get("base_path", "?")

        progress_monitor = progress_monitor_factory.upload_monitor(
            source=f"{total_file_count} files from {base_path}",
            size=total_file_count,
            kwargs={"unit": "file"},
        )

        def on_done_cb():
            progress_monitor.update(1)

        subscriber = DynamicCallbackSubscriber(on_done_cb=on_done_cb)

        with transfer_manager:
            for record, local_path in file_generator:
                local_path.parent.mkdir(parents=True, exist_ok=True)
                transfer_manager.download(
                    record.bucket,
                    record.key,
                    str(local_path),
                    subscribers=[subscriber],
                )
        progress_monitor.close()



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

import pydantic

from ...notifications import (
    NotificationChannel,
    NotificationType,
)


class UserRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a user.
    """

    user_id: str
    is_service_user: bool = False
    """
    A service user can be used to perform actions on behalf of customers.
    For example, a service user can be associated with a Trigger,
    which will then invoke its corresponding Action as the service user.
    """

    is_system_user: Optional[bool] = False

    name: Optional[str] = None
    picture_url: Optional[str] = None
    notification_channels_enabled: dict[NotificationChannel, bool] = pydantic.Field(
        default_factory=dict
    )
    notification_types_enabled: dict[NotificationType, bool] = pydantic.Field(
        default_factory=dict
    )

    def is_email_notifications_enabled(self) -> bool:
        return self.notification_channels_enabled.get(NotificationChannel.Email, False)

    def is_comment_mentions_enabled(self) -> bool:
        return self.notification_types_enabled.get(
            NotificationType.CommentMention, False
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing

from ..config import RobotoConfig
from ..http import RobotoClient
from ..waiters import wait_for
from .api import (
    QueryRecord,
    QueryStatus,
    QueryTarget,
    SubmitRoboqlQueryRequest,
    SubmitStructuredQueryRequest,
)
from .specification import QuerySpecification

RoboQLQuery: typing.TypeAlias = str
Query: typing.TypeAlias = typing.Union[RoboQLQuery, QuerySpecification]


class QueryClient:
    """
    A low-level Roboto query client. Prefer :py:class:`~roboto.query.roboto_search.RobotoSearch` for a simpler,
    more curated query interface.
    """

    __owner_org_id: typing.Optional[str] = None
    __roboto_client: RobotoClient

    def __init__(
        self,
        roboto_client: typing.Optional[RobotoClient] = None,
        owner_org_id: typing.Optional[str] = None,
        roboto_profile: typing.Optional[str] = None,
    ):
        self.__owner_org_id = owner_org_id

        if roboto_client is None:
            roboto_config = RobotoConfig.from_env(profile_override=roboto_profile)
            roboto_client = RobotoClient.from_config(roboto_config)

        self.__roboto_client = roboto_client

    @property
    def roboto_client(self) -> RobotoClient:
        return self.__roboto_client

    def is_query_completed(
        self, query_id: str, owner_org_id: typing.Optional[str] = None
    ) -> bool:
        return (
            self.get_query_record(query_id, owner_org_id).status
            == QueryStatus.Completed
        )

    def get_query_record(
        self, query_id: str, owner_org_id: typing.Optional[str] = None
    ) -> QueryRecord:
        response = self.__roboto_client.get(
            f"v1/query/id/{query_id}", owner_org_id=owner_org_id or self.__owner_org_id
        )
        return response.to_record(QueryRecord)

    def get_query_results(
        self, query_id: str, owner_org_id: typing.Optional[str] = None
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        query_params: dict[str, str] = {}
        while True:
            response = self.__roboto_client.get(
                f"v1/query/id/{query_id}/results",
                query=query_params,
                owner_org_id=owner_org_id or self.__owner_org_id,
            )
            response_data = response.to_dict(json_path=["data"])
            for record in response_data["items"]:
                yield record

            if "next_token" in response_data and response_data["next_token"]:
                query_params["page_token"] = response_data["next_token"]
            else:
                break

    def submit_query(
        self,
        query: typing.Optional[Query],
        target: QueryTarget,
        timeout_seconds: float,
        owner_org_id: typing.Optional[str] = None,
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        if query is None:
            query = QuerySpecification()

        if isinstance(query, QuerySpecification):
            return self.submit_structured_and_await_results(
                request=SubmitStructuredQueryRequest(query=query, target=target),
                timeout_seconds=timeout_seconds,
                owner_org_id=owner_org_id,
            )

        return self.submit_roboql_and_await_results(
            request=SubmitRoboqlQueryRequest(query=query, target=target),
            timeout_seconds=timeout_seconds,
            owner_org_id=owner_org_id,
        )

    def submit_roboql(
        self,
        request: SubmitRoboqlQueryRequest,
        owner_org_id: typing.Optional[str] = None,
    ) -> QueryRecord:
        response = self.__roboto_client.post(
            "v1/query/submit/roboql",
            data=request,
            owner_org_id=owner_org_id or self.__owner_org_id,
        )
        return response.to_record(QueryRecord)

    def submit_roboql_and_await_results(
        self,
        request: SubmitRoboqlQueryRequest,
        timeout_seconds: float,
        owner_org_id: typing.Optional[str] = None,
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        query = self.submit_roboql(request, owner_org_id)

        wait_for(
            self.is_query_completed,
            args=[query.query_id],
            timeout=timeout_seconds,
            interval=2,
        )

        yield from self.get_query_results(query.query_id)

    def submit_structured(
        self,
        request: SubmitStructuredQueryRequest,
        owner_org_id: typing.Optional[str] = None,
    ) -> QueryRecord:
        response = self.__roboto_client.post(
            "v1/query/submit/structured",
            data=request,
            owner_org_id=owner_org_id or self.__owner_org_id,
        )
        return response.to_record(QueryRecord)

    def submit_structured_and_await_results(
        self,
        request: SubmitStructuredQueryRequest,
        timeout_seconds: float,
        owner_org_id: typing.Optional[str] = None,
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        query = self.submit_structured(request, owner_org_id)

        wait_for(
            self.is_query_completed,
            args=[query.query_id],
            timeout=timeout_seconds,
            interval=2,
        )

        yield from self.get_query_results(query.query_id)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum

import pydantic

from ..exceptions import RobotoConditionException


class NotificationType(str, Enum):
    """Notification type enum"""

    CommentMention = "comment_mention"
    # CommentOnThread = "comment_on_thread"
    # CommentOnAuthoredAction = "comment_on_authored_action"
    # etc...


class NotificationChannel(str, Enum):
    """Notification channel enum"""

    Email = "email"
    WebUi = "web_ui"
    # Slack = "slack"
    # SMS = "sms"
    # ApplePush = "apple_push"
    # AndroidPush = "android_push"
    # etc...


class ReadStatus(str, Enum):
    """Notification read status enum"""

    Unread = "unread"
    Read = "read"


# https://docs.aws.amazon.com/ses/latest/dg/event-publishing-retrieving-sns-examples.html#event-publishing-retrieving-sns-subscription
class EmailLifecycleStatus(str, Enum):
    """Notification email lifecycle status enum"""

    Initiated = "Initiated"  # for Roboto use only, this is not an SES status
    Send = "Send"
    Failed = "Failed"  # for Roboto use only, this is not an SES status
    Reject = "Reject"
    Bounce = "Bounce"
    Complaint = "Complaint"
    Delivery = "Delivery"
    Open = "Open"
    Click = "Click"
    RenderingFailure = "RenderingFailure"
    DeliveryDelay = "DeliveryDelay"
    Subscription = "Subscription"


class WebUiLifecycleStatus(str, Enum):
    """Notification web ui status enum"""

    Acknowledged = "Acknowledged"


class LifecycleStatusValidator:
    @pydantic.field_validator("lifecycle_status", mode="before")
    @classmethod
    def check_lifecycle_status(cls, value):
        for channel, status in value.items():
            if channel == NotificationChannel.Email:
                try:
                    EmailLifecycleStatus(status)
                except ValueError:
                    raise RobotoConditionException(
                        f"Invalid status: {status}. Must be one of {EmailLifecycleStatus.__members__.keys()}"
                    )
            elif channel == NotificationChannel.WebUi:
                try:
                    WebUiLifecycleStatus(status)
                except ValueError:
                    raise RobotoConditionException(
                        f"Invalid status: {status}. Must be one of {WebUiLifecycleStatus.__members__.keys()}"
                    )

        return value



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Any, Optional

import pydantic

from .action_record import (
    ComputeRequirements,
    ContainerParameters,
)
from .invocation_record import (
    InvocationDataSourceType,
    InvocationSource,
    InvocationStatus,
)


class CreateInvocationRequest(pydantic.BaseModel):
    """
    Request payload to create a new invocation
    """

    compute_requirement_overrides: Optional[ComputeRequirements] = None
    container_parameter_overrides: Optional[ContainerParameters] = None
    data_source_id: str
    data_source_type: InvocationDataSourceType
    idempotency_id: Optional[str] = None
    input_data: list[str]
    invocation_source: InvocationSource
    invocation_source_id: Optional[str] = None
    parameter_values: Optional[dict[str, Any]] = None
    timeout: Optional[int] = None


class SetContainerInfoRequest(pydantic.BaseModel):
    """
    Set container info for an invocation.

    Called from the `monitor` process once the action image has been pulled.
    """

    image_digest: str


class SetLogsLocationRequest(pydantic.BaseModel):
    """
    Set location where invocation logs are saved to files.

    Called from the `invocation-scheduling-service` once the invocation has been scheduled.
    """

    bucket: str
    prefix: str


class UpdateInvocationStatus(pydantic.BaseModel):
    """
    Request payload to update invocation status
    """

    status: InvocationStatus
    detail: str
